__all__ = [
    "ConstantBiasParameterInitializer",
    "bias_constant_",
    "recursive_bias_constant_",
    "recursive_constant_",
]

import logging
from typing import Union

from torch import nn

from gravitorch.engines.base import BaseEngine
from gravitorch.utils.parameter_initializers.base import BaseDefaultParameterInitializer

logger = logging.getLogger(__name__)


class ConstantBiasParameterInitializer(BaseDefaultParameterInitializer):
    r"""Implements a model parameter initializer with the Xavier Normal or
    uniform strategy.

    Args:
        value (float): Specifies the value to initialize the
            parameters with.
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``
        show_stats (bool, optional): If ``True``, the parameter
            statistics are shown at the end of the initialization.
            Default: ``True``
    """

    def __init__(
        self,
        value: Union[int, float] = 0,
        learnable_only: bool = True,
        log_info: bool = False,
        show_stats: bool = True,
    ):
        super().__init__(show_stats=show_stats)
        self._value = float(value)
        self._learnable_only = bool(learnable_only)
        self._log_info = bool(log_info)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(value={self._value}, "
            f"learnable_only={self._learnable_only}, log_info={self._log_info})"
        )

    def _initialize(self, engine: BaseEngine) -> None:
        if self._learnable_only:
            logger.info(f"Initializing the learnable biases to {self._value}...")
        else:
            logger.info(f"Initializing all biases to {self._value}...")
        recursive_bias_constant_(
            module=engine.model,
            value=self._value,
            learnable_only=self._learnable_only,
            log_info=self._log_info,
        )


def bias_constant_(module: nn.Module, value: float) -> None:
    r"""Initializes the biases of some modules to ``value``.

    This function initializes the biases of the following modules:

        - ``torch.nn.Linear``
        - ``torch.nn.Bilinear``
        - ``torch.nn.ConvNd``
        - ``torch.nn.LayerNorm``
        - ``torch.nn.GroupNorm``
        - ``torch.nn.BatchNormNd``

    Args:
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        value (float): Specifies the value to initialize the
            parameters with.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.parameter_initializers import bias_constant_
        >>> from torch import nn
        >>> layer = nn.Linear(4, 6)
        >>> bias_constant_(layer, value=2)
        >>> layer.weight
        Parameter containing:
        tensor([[ 0.1179, -0.2250, -0.0784, -0.0855],
                [-0.3958, -0.4466, -0.3180, -0.0501],
                [-0.3657, -0.3836, -0.2601,  0.3431],
                [-0.2073, -0.2916,  0.1014,  0.4463],
                [ 0.1758,  0.1229,  0.2563,  0.2391],
                [ 0.2957,  0.1145,  0.3906, -0.2804]], requires_grad=True)
        >>> layer.bias
        Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
    """
    layers_with_bias = (
        nn.Linear,
        nn.Bilinear,
        nn.Conv1d,
        nn.Conv2d,
        nn.Conv3d,
        nn.LayerNorm,
        nn.GroupNorm,
        nn.BatchNorm1d,
        nn.BatchNorm2d,
        nn.BatchNorm3d,
    )
    if isinstance(module, layers_with_bias) and module.bias is not None:
        nn.init.constant_(module.bias, value)
    elif isinstance(module, nn.MultiheadAttention):
        for bias in [module.in_proj_bias, module.bias_k, module.bias_v, module.out_proj.bias]:
            if bias is not None:
                nn.init.constant_(bias, value)


def recursive_bias_constant_(
    module: nn.Module,
    value: Union[int, float],
    learnable_only: bool = True,
    log_info: bool = False,
) -> None:
    r"""Recursively initialize the biases with ``value``.

    To identify the biases, this function looks at if ``'bias'`` is
    in the parameter name.

    Args:
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        value (float): Specifies the value to initialize the
            parameters with.
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.parameter_initializers import recursive_bias_constant_
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> recursive_bias_constant_(net, 2)
        >>> for key, param in net.named_parameters():
        ...     print(key, param)
        0.weight Parameter containing:
        tensor([[ 0.1911,  0.2128, -0.3738,  0.3777],
                [ 0.3394,  0.1795, -0.1418,  0.1076],
                [-0.2050, -0.0029, -0.2555,  0.2612],
                [ 0.1260, -0.3503,  0.1254, -0.3669],
                [ 0.1261, -0.1652,  0.1695,  0.3879],
                [-0.0405,  0.0372,  0.0515, -0.4335]], requires_grad=True)
        0.bias Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        2.weight Parameter containing:
        tensor([1., 1., 1., 1., 1., 1.], requires_grad=True)
        2.bias Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        3.weight Parameter containing:
        tensor([[-0.2702,  0.2759, -0.2408,  0.1202, -0.1770,  0.0223]],
               requires_grad=True)
        3.bias Parameter containing:
        tensor([2.], requires_grad=True)
    """
    for name, params in module.named_parameters():
        if "bias" in name and (not learnable_only or learnable_only and params.requires_grad):
            nn.init.constant_(params.data, value)
            if log_info:
                logger.info(f"Initialize '{name}' to {value} | shape={params.shape}")


def recursive_constant_(
    module: nn.Module, value: Union[int, float], learnable_only: bool = True
) -> None:
    r"""Initialize the parameters of the module with ``value``.

    Args:
        module (``torch.nn.Module``): Specifies the module with the
            parameters to initialize.
        value (int, float): Specifies the value to initialize the
            parameters with.
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.parameter_initializers import recursive_constant_
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> recursive_constant_(net, 2)
        >>> for key, param in net.named_parameters():
        ...     print(key, param)
        0.weight Parameter containing:
        tensor([[2., 2., 2., 2.],
                [2., 2., 2., 2.],
                [2., 2., 2., 2.],
                [2., 2., 2., 2.],
                [2., 2., 2., 2.],
                [2., 2., 2., 2.]], requires_grad=True)
        0.bias Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        2.weight Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        2.bias Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        3.weight Parameter containing:
        tensor([[2., 2., 2., 2., 2., 2.]], requires_grad=True)
        3.bias Parameter containing:
        tensor([2.], requires_grad=True)
    """
    for params in module.parameters():
        if not learnable_only or learnable_only and params.requires_grad:
            nn.init.constant_(params.data, value)



__all__ = ["BaseArtifact", "JSONArtifact", "PickleArtifact", "PyTorchArtifact", "TextArtifact"]

from gravitorch.utils.artifacts.base import BaseArtifact
from gravitorch.utils.artifacts.jsona import JSONArtifact
from gravitorch.utils.artifacts.picklea import PickleArtifact
from gravitorch.utils.artifacts.pytorcha import PyTorchArtifact
from gravitorch.utils.artifacts.text import TextArtifact



__all__ = [
    "ddp_partitions",
    "even_partitions",
    "fixed_size_partitions",
    "select_partition_by_rank",
    "split_into_two_partitions",
]

import math
from collections.abc import Sequence
from typing import TypeVar

import torch

from gravitorch import distributed as dist
from gravitorch.utils.seed import get_torch_generator

T = TypeVar("T")


def ddp_partitions(items: Sequence[T], partition_size: int) -> list[Sequence[T]]:
    r"""Creates non-overlapping partitions of a sequence of items in a
    Distributed Data Parallel (DDP) setting.

    All the partitions have the same number of items. The number of
    partitions is the world size of the distributed system.

    Args:
        items (sequence): Specifies the sequence to partition. The
            length should be greater or equal to
            ``partition_size * world_size``.
        partition_size (int): Specifies the partition size i.e. the
            number of items in each partition.

    Returns:
        list: A partition of the items.

    Raises:
        ValueError if ``items`` has less than `
            `partition_size * world_size`` items.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.partitioning import ddp_partitions
        # World size 1
        >>> ddp_partitions([1, 2, 3, 4, 5, 6, 7, 8], partition_size=2)
        [[1, 2]]
        >>> ddp_partitions([1, 2, 3, 4, 5, 6, 7, 8], partition_size=3)
        [[1, 2, 3]]
        # World size 2
        >>> ddp_partitions([1, 2, 3, 4, 5, 6, 7, 8], partition_size=2)
        [[1, 2], [3, 4]]
        >>> ddp_partitions([1, 2, 3, 4, 5, 6, 7, 8], partition_size=3)
        [[1, 2, 3], [4, 5, 6]]
    """
    world_size = dist.get_world_size()
    if len(items) < partition_size * world_size:
        raise ValueError(
            f"Incorrect `partition_size` {partition_size:,}. The partition size should be lower "
            f"than {len(items) // world_size:,}"
        )
    return [items[i * partition_size : (i + 1) * partition_size] for i in range(world_size)]


def even_partitions(
    items: Sequence[T], num_partitions: int, drop_remainder: bool = False
) -> list[Sequence[T]]:
    r"""Creates non-overlapping partitions of a sequence of items.

    The items are distributed evenly across partitions, starting by
    the first.

    Args:
        items (sequence): Specifies the sequence to partition.
        num_partitions (int): Specifies the number of partitions.
        drop_remainder (bool, optional): If ``True``, it drops the
            last items if the number of items is not evenly divisible
            by ``num_partitions``.

    Returns:
        list: A partition of the items.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.partitioning import even_partitions
        >>> even_partitions([1, 2, 3, 4, 5, 6, 7, 8], num_partitions=3)
        [[1, 2, 3], [4, 5, 6], [7, 8]]
        >>> even_partitions([1, 2, 3, 4, 5, 6, 7, 8], num_partitions=3, drop_remainder=True)
        [[1, 2], [3, 4], [5, 6]]
    """
    rounding_op = math.floor if drop_remainder else math.ceil
    partition_size = rounding_op(len(items) / num_partitions)
    return [items[i * partition_size : (i + 1) * partition_size] for i in range(num_partitions)]


def fixed_size_partitions(
    items: Sequence[T], partition_size: int, drop_last: bool = False
) -> list[Sequence[T]]:
    r"""Creates non-overlapping partitions of a sequence of items.

    The items are distributed across partitions, starting by the first.

    Args:
        items (sequence): Specifies the sequence to partition.
        partition_size (int): Specifies the number of items per
            partitions.
        drop_last (bool, optional): If ``True``, it drops the last
            items if the number of items is not evenly divisible by
            ``partition_size``.

    Returns:
        list: A partition of the items.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.partitioning import fixed_size_partitions
        >>> fixed_size_partitions([1, 2, 3, 4, 5, 6, 7, 8], partition_size=3)
        [[1, 2, 3], [4, 5, 6], [7, 8]]
        >>> fixed_size_partitions([1, 2, 3, 4, 5, 6, 7, 8], partition_size=3, drop_last=True)
        [[1, 2, 3], [4, 5, 6]]
    """
    rounding_op = math.floor if drop_last else math.ceil
    num_partitions = rounding_op(len(items) / partition_size)
    return [items[i * partition_size : (i + 1) * partition_size] for i in range(num_partitions)]


def select_partition_by_rank(partitions: Sequence[T]) -> T:
    r"""Selects a partition by using the current distributed rank.

    Args:
        partitions: Specifies the sequence of partitions

    Returns:
        The selected partition for the current distributed rank.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.partitioning import select_partition_by_rank
        # Rank 0
        >>> select_partition_by_rank([1, 2, 3, 4])
        1
        # Rank 1
        >>> select_partition_by_rank([1, 2, 3, 4])
        2
    """
    world_size = dist.get_world_size()
    if len(partitions) < world_size:
        raise ValueError(
            f"Incorrect number of partitions. The number of partitions ({len(partitions)}) "
            f"has to be greater or equal to the distributed world size ({world_size})"
        )
    return partitions[dist.get_rank()]


def split_into_two_partitions(
    items: Sequence[T], first_partition_ratio: float = 0.8, random_seed: int = 4979957926458772275
) -> tuple[tuple[T, ...], tuple[T, ...]]:
    r"""Splits the items into two partitions.

    This function can be used to create train/val or train/test
    dataset splits.

    Args:
        items (sequence): Specifies the items to split into two
            partitions.
        first_partition_ratio (float, optional): Specifies the ratio
            of the first partition. ``0.8`` means that 80% of the
            items will be split into the first partition.
        random_seed (int, optional): Specifies the random seed used
            to create the partitions. Default: ``4979957926458772275``

    Example usage:

    .. code-block:: python

        >>> from gravitorch.utils.partitioning import split_into_two_partitions
        >>> first, second = split_into_two_partitions([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        >>> first
        (2, 7, 4, 1, 6, 5, 3, 8)
        >>> second
        (9, 0)
    """
    permutation = torch.randperm(len(items), generator=get_torch_generator(random_seed))
    num_items_first_partition = int(first_partition_ratio * len(items))
    return (
        tuple(items[i] for i in permutation[:num_items_first_partition]),
        tuple(items[i] for i in permutation[num_items_first_partition:]),
    )



r"""This module defines some utility functions for distributed experiments."""

__all__ = [
    "conditional_evaluation",
    "has_slurm_distributed_env_vars",
    "has_torch_distributed_env_vars",
    "is_slurm_job",
    "is_distributed_ready",
    "show_all_slurm_env_vars",
    "show_distributed_context_info",
    "show_distributed_env_vars",
    "show_slurm_env_vars",
    "show_torch_distributed_env_vars",
]

import logging
import os
from typing import Callable

from gravitorch.distributed import comm as dist
from gravitorch.distributed._constants import (
    CUDA_VISIBLE_DEVICES,
    SLURM_DISTRIBUTED_ENV_VARS,
    SLURM_JOB_ID,
    SLURM_NTASKS,
    TORCH_DISTRIBUTED_ENV_VARS,
)
from gravitorch.utils.format import to_pretty_dict_str

logger = logging.getLogger(__name__)


def is_slurm_job() -> bool:
    r"""Indicates if the current process is connected to a SLURM job.

    A process is considered connected to a SLURM job if the
    environment variable ``SLURM_JOB_ID`` exists.

    Returns:
        bool: ``True`` if the current process is connected to a SLURM
            job, otherwise ``False``.
    """
    return SLURM_JOB_ID in os.environ


def is_distributed_ready() -> bool:
    r"""Indicates if the distributed context is ready to be activated or not.

    Returns:
        bool: ``True`` if the distributed context should be activated,
            otherwise ``False``.
    """
    if is_slurm_job():
        return has_slurm_distributed_env_vars() and int(os.environ.get(SLURM_NTASKS, 0)) > 1
    return has_torch_distributed_env_vars()


def has_torch_distributed_env_vars() -> bool:
    r"""Indicates if the environment variables for the native PyTorch
    distributed backends are set or not.

    It assumes that the distributed processes were launched with
    ``torch.distributed.run``
    (https://github.com/pytorch/pytorch/blob/master/torch/distributed/run.py)
    or any other tools that setup the same environment variables.

    Returns:
        bool: ``True`` if all the environment variables are set,
            otherwise ``False``.
    """
    for env_var in TORCH_DISTRIBUTED_ENV_VARS:
        if env_var not in os.environ:
            return False
    return True


def has_slurm_distributed_env_vars() -> bool:
    r"""Indicates if the environment variables required to initialize the native
    PyTorch distributed backends are set or not.

    Returns:
        bool: ``True`` if all the environment variables are set,
            otherwise ``False``.
    """
    for env_var in SLURM_DISTRIBUTED_ENV_VARS:
        if env_var not in os.environ:
            return False
    return True


def show_distributed_env_vars() -> None:
    r"""Shows the value of some environment variables used to set up the
    distributed context."""
    show_torch_distributed_env_vars()
    if is_slurm_job():
        show_slurm_env_vars()
        show_all_slurm_env_vars()


def show_torch_distributed_env_vars() -> None:
    r"""Shows the value of the environment variables used to set up the PyTorch
    distributed context.

    This function shows the value of the following environment
    variables:

        - WORLD_SIZE: The total number of processes, so that the
            master knows how many workers to wait for.
        - RANK: The global rank of the process. Rank is a unique
            identifier assigned to each process within a distributed
            process group. They are always consecutive integers
            ranging from 0 to ``WORLD_SIZE``.
        - LOCAL_RANK: The local rank of the process. This value is
            not unique globally, but it is unique per node.
        - MASTER_ADDR: IP address (or hostname) of the machine that
            will host the process with rank 0.
        - MASTER_PORT: The master node (rank 0)'s free port that
            needs to be used for communication during distributed
            training.
        - The remaining environment variables can be found in the
            documentation of ``torch.distributed.run``
            https://github.com/pytorch/pytorch/blob/master/torch/distributed/run.py
    """
    names = TORCH_DISTRIBUTED_ENV_VARS + (CUDA_VISIBLE_DEVICES,)
    env_vars = {env_var: os.environ.get(env_var, "<NOT SET>") for env_var in names}
    logger.info(
        "PyTorch environment variables for distributed training:\n"
        f"{to_pretty_dict_str(env_vars, sorted_keys=True, indent=2)}\n"
    )


def show_all_slurm_env_vars() -> None:
    r"""Shows the value of the all the SLURM environment variables."""
    env_vars = {key: value for key, value in os.environ.items() if key.startswith("SLURM_")}
    logger.info(
        "All SLURM environment variables:\n"
        f"{to_pretty_dict_str(env_vars, sorted_keys=True, indent=2)}\n"
    )


def show_slurm_env_vars() -> None:
    r"""Shows the value of some SLURM environment variables."""
    env_vars = {
        env_var: os.environ.get(env_var, "<NOT SET>") for env_var in SLURM_DISTRIBUTED_ENV_VARS
    }
    logger.info(
        "SLURM environment variables:\n"
        f"{to_pretty_dict_str(env_vars, sorted_keys=True, indent=2)}\n"
    )


def show_distributed_context_info() -> None:
    r"""Shows some information about the distributed context.

    This information is useful to verify that each process is well
    configured.
    """
    logger.info(
        f"Config of the distributed context:\n"
        f"  model_name         : {dist.model_name()}\n"
        f"  backend            : {dist.backend()}\n"
        f"  device             : {dist.device()}\n"
        f"  hostname           : {dist.hostname()}\n"
        f"  get_world_size     : {dist.get_world_size()}\n"
        f"  get_rank           : {dist.get_rank()}\n"
        f"  get_local_rank     : {dist.get_local_rank()}\n"
        f"  is_main_process    : {dist.is_main_process()}\n"
        f"  get_nnodes         : {dist.get_nnodes()}\n"
        f"  get_nproc_per_node : {dist.get_nproc_per_node()}\n"
        f"  get_node_rank      : {dist.get_node_rank()}\n"
    )


def conditional_evaluation(only_main_process: bool, callable: Callable, *args, **kwargs) -> None:
    r"""Evaluates or not the callable based on the distributed context.

    Args:
        only_main_process (bool, optional): If ``True``, only the
            main process evaluates the callable otherwise
            all the processes evaluate the callable.
        callable (Callable): Specifies the callable to evaluate.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Example:

    .. code-block:: python

        >>> from gravitorch.distributed.utils import conditional_evaluation
        >>> conditional_evaluation(False, print, "hello")
        hello
        # If main process
        >>> conditional_evaluation(True, print, "hello")
        hello
        # If non main process
        >>> conditional_evaluation(True, print, "hello")
    """
    if only_main_process and not dist.is_main_process():
        return
    callable(*args, **kwargs)



__all__ = ["PyTorchArtifact"]

from pathlib import Path
from typing import Any

from gravitorch.utils.artifacts.base import BaseArtifact
from gravitorch.utils.io import save_pytorch


class PyTorchArtifact(BaseArtifact):
    r"""Implements a PyTorch artifact.

    Args:
        tag (str): Specifies the artifact tag. The tag is used to
            define the PyTorch filename.
        data: Specifies the data to save in the PyTorch artifact. The
            data should be PyTorch compatible.
    """

    def __init__(self, tag: str, data: Any):
        self._tag = str(tag)
        self._data = data

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(tag={self._tag})"

    def create(self, path: Path) -> None:
        save_pytorch(to_save=self._data, path=path.joinpath(f"{self._tag}.pt"))



r"""This module defines some collators using padding to deal with variable size
data."""

__all__ = ["PaddedSequenceCollator", "DictPaddedSequenceCollator"]

from collections.abc import Hashable, Sequence

from torch.nn.utils.rnn import pad_sequence
from torch.utils.data.dataloader import default_collate

from gravitorch import constants as ct
from gravitorch.data.dataloaders.collators.base import BaseCollator


class PaddedSequenceCollator(BaseCollator[tuple[dict, dict], dict]):
    r"""Implements a collate function to create zero padded sequences.

    Each sequence can have a different length. The missing "values"
    are filled with 0 (default value). The examples have to be
    formatted in a specified format. The example should be a tuple
    with two items. The input should be a tuple with two items. The
    first item is a dict containing the fixed size inputs i.e. the
    input that have the same size for each example of the mini-batch.
    The first item needs to contain the length of the sequence
    because this information is used to sort the examples in the
    batch by descending order of sequence length. The length of the
    sequence is indicated by the key ``'length'``. The first item is
    combined by the default PyTorch collate function. The second item
    is a dict containing all the inputs that have variable size. This
    item is combined by using packed sequences. This collator does
    not work if all the sequences are empty. Note that every empty
    sequence if removed.

    Args:
        length_key (hashable, optional): Specifies the key with the
            length of each example. Default: ``'length'``
        batch_first (bool, optional): Specifies if the first dimension
            is the batch size or the sequence length.
            Default: ``False``
        padding_value (float, optional): value for padded elements.
            Default: ``0.0``
    """

    def __init__(
        self,
        length_key: Hashable = ct.LENGTH,
        batch_first: bool = False,
        padding_value: float = 0.0,
    ):
        self._length_key = length_key
        self._batch_first = batch_first
        self._padding_value = padding_value

    def __call__(self, data: list[tuple[dict, dict]]) -> dict:
        r"""Creates a batch of packed sequence examples given a list of
        examples.

        Args:
            data (list): The list of examples.

        Returns:
            dict: The generated mini-batch.
        """
        # Sort the examples by the length of their sequence.
        data.sort(key=lambda x: x[0][self._length_key], reverse=True)

        fixed_size_data, variable_size_data = zip(*data)
        # Create the batch with the fixed size inputs and remove examples with empty sequences.
        batch = default_collate(fixed_size_data)
        lengths = batch[self._length_key]
        num_seq = (lengths > 0).sum().item()

        # Remove the empty sequences.
        for key, value in batch.items():
            batch[key] = value[:num_seq]

        # Create the zero padded sequences.
        for key in variable_size_data[0].keys():
            batch[key] = pad_sequence(
                [variable_size_data[i][key] for i in range(num_seq)],
                batch_first=self._batch_first,
                padding_value=self._padding_value,
            )

        return batch

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}("
            f"length_key={self._length_key}, "
            f"batch_first={self._batch_first}, "
            f"padding_value={self._padding_value})"
        )


class DictPaddedSequenceCollator(BaseCollator[dict, dict]):
    r"""Implements a collator to create batch of padded sequences because the
    default PyTorch collate function does not work with sequences of different
    lengths.

    With this collator, each example in the batch can have a different
    length. Each example has to be a dict, and you need to specify the
    list of keys that will be aggregated by using the function
    ``pad_sequence`` of PyTorch.

    .. note::

        This collator will not raise an error if you specify a key
        that is not in the keys of the batch.

    Args:
        keys_to_pad (list): Specifies the list of keys to pack. The
            first key is used to sort the examples by length.
        batch_first (bool, optional): Specifies if the first dimension
            is the batch size or the sequence length.
            Default: ``False``
        padding_value (float, optional): value for padded elements.
            Default: ``0.``
    """

    def __init__(
        self,
        keys_to_pad: Sequence[Hashable],
        batch_first: bool = False,
        padding_value: float = 0.0,
    ):
        self._keys_to_pad = tuple(keys_to_pad)
        self._batch_first = batch_first
        self._padding_value = padding_value

    def __call__(self, data: Sequence[dict]) -> dict:
        r"""Creates a batch of packed sequence examples given a list of
        examples.

        Args:
            data: The sequence of examples.

        Returns:
            dict: The generated mini-batch.
        """
        # Remove the empty sequences.
        data = [example for example in data if example[self._keys_to_pad[0]].shape[0] > 0]
        num_seq = len(data)

        # Sort the examples by the length of their sequence.
        data.sort(key=lambda x: x[self._keys_to_pad[0]].shape[0], reverse=True)

        # Create the batch of examples.
        batch = {}
        for key in data[0].keys():
            examples = [data[i][key] for i in range(num_seq)]
            if key in self._keys_to_pad:
                batch[key] = pad_sequence(
                    examples, batch_first=self._batch_first, padding_value=self._padding_value
                )
            else:
                batch[key] = default_collate(examples)

        return batch

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}("
            f"keys_to_pack={self._keys_to_pad}, "
            f"batch_first={self._batch_first}, "
            f"padding_value={self._padding_value})"
        )



__all__ = ["ZeroRedundancyOptimizerCreator"]

import logging
from typing import Optional

from objectory import OBJECT_TARGET
from torch.distributed.optim import ZeroRedundancyOptimizer
from torch.nn import Module
from tornado.util import import_object

from gravitorch import constants as ct
from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.engines.base import BaseEngine
from gravitorch.handlers import ConsolidateOptimizerState
from gravitorch.utils.format import str_add_indent, to_pretty_json_str

logger = logging.getLogger(__name__)


class ZeroRedundancyOptimizerCreator(BaseOptimizerCreator):
    r"""Implements a zero redundancy optimizer (ZeRO) creator.

    Documentation of ``torch.distributed.optim.ZeroRedundancyOptimizer``:
    https://pytorch.org/docs/stable/distributed.optim.html#torch.distributed.optim.ZeroRedundancyOptimizer

    This creator wraps an arbitrary ``torch.optim.Optimizer``
    optimizer and shards its state as described by ZeRO
    (https://arxiv.org/abs/1910.02054).
    Note ZeRO works only for distributed training.

    Args:
        optimizer_config (dict): Specifies the optimizer
            configuration. The dictionary must have a key
            ``'_target_`` that indicates the optimizer to shard. The
            other keys are the arguments of the optimizer to shard.
        zero_kwargs (dict): Specifies some keyword arguments used to
            instantiate the
            ``torch.distributed.optim.ZeroRedundancyOptimizer``.
            Please read the documentation of
            ``torch.distributed.optim.ZeroRedundancyOptimizer`` to see
            the possible options. Note that it is not possible to set
            ``params`` and ``optimizer_class`` with a keyword argument.
        add_module_to_engine (bool, optional): If ``True``, the
            optimizer is added to the engine state, so the optimizer
            state is stored when the engine creates a checkpoint.
            Default: ``True``
        attach_handler (bool, optional): If ``True``, a handler is
            attached to the engine to consolidate the ZeRO optimizer
            state at the end of each training epoch. Consolidate the
            optimizer state dict is important to export the optimizer
            state dict. If ``False``, no handler is attached.
            Default: ``True``
    """

    def __init__(
        self,
        optimizer_config: dict,
        zero_kwargs: Optional[dict] = None,
        add_module_to_engine: bool = True,
        attach_handler: bool = True,
    ):
        self._optimizer_class = import_object(optimizer_config.pop(OBJECT_TARGET))
        self._optimizer_config = optimizer_config
        self._zero_kwargs = zero_kwargs or {}
        self._add_module_to_engine = bool(add_module_to_engine)
        self._attach_handler = bool(attach_handler)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  optimizer_class={self._optimizer_class},\n"
            f"  optimizer_config={str_add_indent(to_pretty_json_str(self._optimizer_config))},\n"
            f"  zero_kwargs={str_add_indent(to_pretty_json_str(self._zero_kwargs))},\n"
            f"  add_module_to_engine={self._add_module_to_engine},\n"
            f"  attach_handler={self._attach_handler},\n"
            ")"
        )

    def create(self, engine: BaseEngine, model: Module) -> ZeroRedundancyOptimizer:
        r"""Creates a zero redundancy optimizer (ZeRO).

        This method is responsible to register the event handlers
        associated to the optimizer.

        Args:
            engine (``gravitorch.engines.BaseEngine``): Specifies an
                engine.
            model (``torch.nn.Module``): Specifies a model.

        Returns:
            ``torch.distributed.optim.ZeroRedundancyOptimizer``: The
                created zero redundancy optimizer (ZeRO).
        """
        optimizer = ZeroRedundancyOptimizer(
            params=model.parameters(),
            optimizer_class=self._optimizer_class,
            **self._optimizer_config,
            **self._zero_kwargs,
        )
        logger.info(f"optimizer:\n{optimizer}")
        if self._add_module_to_engine:
            logger.info(f"Adding an optimizer to the engine (key: {ct.OPTIMIZER})...")
            engine.add_module(ct.OPTIMIZER, optimizer)
        if self._attach_handler:
            logger.info("Creating handler to consolidate the optimizer state dict...")
            ConsolidateOptimizerState().attach(engine)
        return optimizer



__all__ = [
    "CpuDevicePlacement",
    "CudaDevicePlacement",
    "ManualDevicePlacement",
    "MpsDevicePlacement",
]

from typing import Any, Union

import torch

from gravitorch.utils import move_to_device
from gravitorch.utils.device_placement.base import BaseDevicePlacement


class ManualDevicePlacement(BaseDevicePlacement):
    r"""Implements a device placement class to send objects on a given device.

    The user is responsible to choose the target device.
    """

    def __init__(self, device: Union[torch.device, str]):
        if isinstance(device, str):
            device = torch.device(device)
        self.device = device

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(device={self.device})"

    def send(self, obj: Any) -> Any:
        r"""Sends the object on a target device.

        Args:
            obj: The object to send to the target device.

        Returns:
            The object on the target device.
        """
        return move_to_device(obj, self.device)


class CpuDevicePlacement(ManualDevicePlacement):
    r"""Implements a device placement class to send objects on a cpu device."""

    def __init__(self):
        super().__init__(torch.device("cpu"))


class CudaDevicePlacement(ManualDevicePlacement):
    r"""Implements a device placement class to send objects on a cuda device.

    Args:
        index (int, optional): Specifies the index of the cuda device.
            Default: ``0``
    """

    def __init__(self, index: int = 0):
        super().__init__(torch.device(type="cuda", index=index))


class MpsDevicePlacement(ManualDevicePlacement):
    r"""Implements a device placement class to send objects on a Metal
    Performance Shaders (MPS) device."""

    def __init__(self):
        super().__init__(torch.device("mps"))



__all__ = ["BaseEpochMetric", "BaseStateEpochMetric"]

import logging
from typing import Optional, Union

from gravitorch import constants as ct
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.models.metrics.base import BaseMetric
from gravitorch.models.metrics.state import BaseState, setup_state
from gravitorch.utils.events import VanillaEventHandler
from gravitorch.utils.exp_trackers import EpochStep

logger = logging.getLogger(__name__)


class BaseEpochMetric(BaseMetric):
    r"""Implements the base class of an epoch-wise metric.

    By default, this metric is reset at the beginning of the epoch and
    the result is computed at the end of the epoch.

    Child classes must implement the following methods:
        - ``forward``
        - ``reset``
        - ``value``

    Args:
        mode (str): Specifies the mode (e.g train or eval).
        name (str): Specifies the name of the metric.
    """

    def __init__(self, mode: str, name: str):
        super().__init__()
        self._mode = str(mode)
        self._name = str(name)
        self._metric_name = f"{self._mode}/{self._name}"

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches current metric to the provided engine.

        This method adds event handlers to the engine.

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        if self._mode == ct.TRAIN:
            reset_event = EngineEvents.TRAIN_EPOCH_STARTED
            value_event = EngineEvents.TRAIN_EPOCH_COMPLETED
        else:
            reset_event = EngineEvents.EVAL_EPOCH_STARTED
            value_event = EngineEvents.EVAL_EPOCH_COMPLETED

        engine.add_event_handler(reset_event, VanillaEventHandler(self.reset))
        engine.add_event_handler(
            value_event, VanillaEventHandler(self.value, handler_kwargs={"engine": engine})
        )


class BaseStateEpochMetric(BaseEpochMetric):
    r"""Defines a base class to implement a metric where a lower value is better
    than a large value.

    Child classes must implement the following method:
        - ``forward``

    Args:
        mode (str): Specifies the mode (e.g ``'train'`` or ``'eval'``).
        name (str): Specifies the name of the metric. The name is used
            to log the metric results.
        state (``BaseState`` or dict): Specifies the metric state or
            its configuration.
    """

    def __init__(self, mode: str, name: str, state: Union[BaseState, dict]):
        super().__init__(mode=mode, name=name)
        self._state = setup_state(state)

    def extra_repr(self) -> str:
        return f"mode={self._mode},\nname={self._name},\nstate={self._state}"

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches current metric to the provided engine.

        This method can be used to:

            - add event handler to the engine
            - set up history trackers

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        super().attach(engine)
        for history in self._state.get_histories(f"{self._metric_name}_"):
            engine.add_history(history)

    def reset(self) -> None:
        r"""Resets the metric."""
        self._state.reset()

    def value(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Evaluates the metric and log the results given all the predictions
        previously seen.

        Args:
            engine (``BaseEngine`` or ``None``, optional): Specifies
                the engine. This argument is required to log the
                results in the engine. Default: ``None``.

        Returns:
             dict: The results of the metric
        """
        results = self._state.value(f"{self._metric_name}_")
        if engine:
            engine.log_metrics(results, EpochStep(engine.epoch))
        return results



r"""This module implements the base class to implement a data summary."""

__all__ = ["BaseDataSummary", "EmptyDataSummaryError"]

from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from objectory import AbstractFactory

T = TypeVar("T")


class BaseDataSummary(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to implement a data summary.

    Note that the data summary only stores a descriptive summary of the
    data, and not the data.
    """

    @abstractmethod
    def add(self, data: T) -> None:
        r"""Adds new data to the summary.

        Args:
            data: Specifies the data to add to the summary.
        """

    @abstractmethod
    def reset(self) -> None:
        r"""Resets the data summary."""

    @abstractmethod
    def summary(self) -> dict:
        r"""Gets a descriptive summary of the data.

        Returns:
            dict: The data descriptive summary.

        Raises:
            ``EmptyDataSummaryError`` is the data summary is empty.
        """


class EmptyDataSummaryError(Exception):
    r"""Raised when the data summary is empty because it is not possible to
    evaluate an empty data summary."""



r"""This module contains the entrypoint to run experiments.

Example usage:

.. code-block::

python -m gravitorch.cli.run -cd=examples/classification
-cn=mlp_classification_sgd engine.state.max_epochs=2  # noqa: E501,B950
"""

__all__ = ["main", "run_cli"]

import logging
from typing import Any

import hydra
from hya import register_resolvers
from omegaconf import DictConfig, OmegaConf

from gravitorch import constants as ct
from gravitorch.runners import setup_runner
from gravitorch.utils.info import log_run_info
from gravitorch.utils.timing import timeblock

logger = logging.getLogger(__name__)


def main(config: dict[str, Any]) -> None:
    r"""Initializes a runner given its configuration and executes its logic.

    Args:
        config (dict): Specifies the dictionary with the configuration
            of the runner. This dictionary has to have a key
            ``'runner'``.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.cli.run import main
        >>> config = {
        ...     '_target_': 'gravitorch.runners.TrainingRunner',
        ...     'engine': {'_target_': 'gravitorch.engines.AlphaEngine', ...},
        ...     'exp_tracker': {'_target_': 'gravitorch.utils.exp_trackers.NoOpExpTracker', ...},
        ... }
        >>> main(config)
    """
    with timeblock("Total time of the run: {time}"):
        logger.info("Creating runner...")
        runner = setup_runner(config[ct.RUNNER])
        logger.info(f"runner:\n{runner}")
        logger.info("Start to execute the logic of the runner")
        runner.run()
        logger.info("End of the run")


@hydra.main(config_path=None, version_base=None)
def run_cli(config: DictConfig) -> None:
    r"""Defines the CLI entrypoint to run an experiment.

    Please check the Hydra dcoumentation to learn how Hydra works:
    https://hydra.cc/

    Args:
        config (``omegaconf.DictConfig``): Specifies the dictionary
            with the configuration of the runner. This dictionary has
            to have a key ``'runner'``.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.cli.run import run_cli
        >>> run_cli()
    """
    log_run_info(config)
    register_resolvers()
    main(OmegaConf.to_container(config, resolve=True))


if __name__ == "__main__":
    run_cli()



__all__ = ["EpochShufflePartitioner"]

from collections.abc import Sequence
from typing import Optional, TypeVar, Union

import torch

from gravitorch.data.partitioners.base import BasePartitioner, setup_partitioner
from gravitorch.engines import BaseEngine
from gravitorch.utils.format import str_add_indent
from gravitorch.utils.seed import get_torch_generator

T = TypeVar("T")


class EpochShufflePartitioner(BasePartitioner[T]):
    r"""Implements a partitioner that shuffles the data before to partition
    them.

    To be reproducible, the shuffling is controlled by the engine
    epoch value and a base random seed. If no engine is provided,
    the epoch value is set to ``0``.

    Args:
        partitioner (``BasePartitioner`` or dict): Specifies a
            partitioner or its configuration.
        random_seed (int, optional): Specifies the base random seed.
            This random seed is added to the engine epoch value to
            define the initial seed of the ``torch.Generator`` object.
            Default: ``7553907118525846636``
    """

    def __init__(
        self, partitioner: Union[BasePartitioner, dict], random_seed: int = 7553907118525846636
    ):
        self._partitioner = setup_partitioner(partitioner)
        self._random_seed = random_seed

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  partitioner={str_add_indent(self._partitioner)},\n"
            f"  random_seed={self._random_seed},\n"
            ")"
        )

    @property
    def partitioner(self) -> BasePartitioner:
        r"""``BasePartitioner``: The partitioner."""
        return self._partitioner

    @property
    def random_seed(self) -> int:
        r"""int: The base random seed."""
        return self._random_seed

    def partition(
        self, items: Sequence[T], engine: Optional[BaseEngine] = None
    ) -> list[Sequence[T]]:
        epoch = engine.epoch if engine is not None else 0
        permutation = torch.randperm(
            len(items), generator=get_torch_generator(self._random_seed + epoch)
        )
        return self._partitioner.partition([items[i] for i in permutation], engine)



__all__ = [
    "AbsoluteError",
    "AbsoluteRelativeError",
    "LogCoshError",
    "NormalizedMeanSquaredError",
    "RootMeanSquaredError",
    "SquaredAsinhError",
    "SquaredError",
    "SquaredLogError",
    "SquaredSymlogError",
    "SymmetricAbsoluteRelativeError",
]

from gravitorch.models.metrics.regression.absolute_error import AbsoluteError
from gravitorch.models.metrics.regression.absolute_relative_error import (
    AbsoluteRelativeError,
    SymmetricAbsoluteRelativeError,
)
from gravitorch.models.metrics.regression.log_cosh_error import LogCoshError
from gravitorch.models.metrics.regression.nmse import NormalizedMeanSquaredError
from gravitorch.models.metrics.regression.squared_error import (
    RootMeanSquaredError,
    SquaredError,
)
from gravitorch.models.metrics.regression.squared_log_error import (
    SquaredAsinhError,
    SquaredLogError,
    SquaredSymlogError,
)



__all__ = ["InverseSquareRootLR"]

import math

from torch.optim import Optimizer
from torch.optim.lr_scheduler import LambdaLR


def inverse_square_root(step: int) -> float:
    r"""Computes the inverse square root.

    Args:
        step (int): Specifies the current step.

    Returns:
        float: The associated factor for the current step.
    """
    return 1.0 / math.sqrt(1 + step)


class InverseSquareRootLR(LambdaLR):
    r"""Implementation of Inverse Square Root LR scheduler.

    Args:
        optimizer (``torch.optim.Optimizer``): Specifies the optimizer
            associated to the LR scheduler.
        last_epoch (int, optional): Specifies the index of last epoch.
            Default: ``-1``
        verbose (bool, optional): If ``True``, prints a message to
            stdout for each update. Default: ``False``.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.lr_schedulers import InverseSquareRootLR
        >>> optimizer = torch.optim.SGD(torch.nn.Linear(4, 6).parameters(), lr=0.01)
        >>> scheduler = InverseSquareRootLR(optimizer)
    """

    def __init__(self, optimizer: Optimizer, last_epoch: int = -1, verbose: bool = False):
        super().__init__(
            optimizer=optimizer,
            lr_lambda=inverse_square_root,
            last_epoch=last_epoch,
            verbose=verbose,
        )



__all__ = ["ImageNetDataSource", "create_train_eval_datasets_v1"]

import logging
from pathlib import Path
from typing import Optional, Union

from torch.utils.data import Dataset

from gravitorch import constants as ct
from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.data.datasets import ImageFolderDataset
from gravitorch.datasources.dataset import DatasetDataSource
from gravitorch.utils.integrations import check_torchvision, is_torchvision_available

if is_torchvision_available():
    from torchvision import transforms
else:
    transforms = None  # pragma: no cover

logger = logging.getLogger(__name__)

IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)


class ImageNetDataSource(DatasetDataSource):
    r"""Implements a data source for the ImageNet dataset."""

    @classmethod
    def create_imagenet_v1(
        cls,
        train_path: Union[Path, str, None],
        eval_path: Union[Path, str, None],
        data_loader_creators: dict[str, Union[BaseDataLoaderCreator, dict, None]],
        input_size: int = 224,
    ) -> "ImageNetDataSource":
        r"""Creates a data source for the ImageNet dataset the training and
        evaluation datasets with their original data augmentations (V1).

        This function initializes the original data augmentations,
        sometimes called V1
        (https://pytorch.org/blog/introducing-torchvision-new-multi-weight-support-api/).
        Note that more aggressive data augmentation can be used to
        improve the performances. See the following article for more
        information:

            Wightman R., Touvron H., Jégou H.
            ResNet strikes back: An improved training procedure in timm.
            http://arxiv.org/pdf/2110.00476

        Args:
            train_path (``pathlib.Path`` or str or ``None``):
                Specifies the path to the training dataset.
                If ``None``, the training dataset is set to ``None``.
            eval_path (``pathlib.Path`` or str or ``None``):
                Specifies the path to the evaluation dataset.
                If ``None``, the evaluation dataset is set to ``None``.
            data_loader_creators: (dict): Specifies the data loader
                creators to initialize. Each key indicates a data
                loader creator name. For example if you want to create
                a data loader for ``'train'`` ID, the dictionary has to
                have a key ``'train'``. The value can be a
                ``BaseDataLoaderCreator`` object, or its configuration,
                or ``None``. ``None`` means a default data loader will
                be created. Each data loader creator takes a
                ``Dataset`` object as input, so you need to specify
                the associated dataset path.
            input_size (int, optional): Specifies the input size for
                the image. Default: ``224``

        Returns:
            ``ImageNetDataSource``: A data source for the ImageNet datasets.
        """
        train_dataset, eval_dataset = create_train_eval_datasets_v1(
            train_path=train_path,
            eval_path=eval_path,
            input_size=input_size,
        )
        datasets = {}
        if train_dataset:
            datasets[ct.TRAIN] = train_dataset
        if eval_dataset:
            datasets[ct.EVAL] = eval_dataset
        return cls(datasets=datasets, data_loader_creators=data_loader_creators)


def create_train_eval_datasets_v1(
    train_path: Union[Path, str, None],
    eval_path: Union[Path, str, None],
    input_size: int = 224,
) -> tuple[Optional[Dataset], Optional[Dataset]]:
    r"""Creates the training and evaluation datasets with the original data
    augmentations (V1).

    This function initializes the original data augmentations,
    sometimes called V1
    (https://pytorch.org/blog/introducing-torchvision-new-multi-weight-support-api/).
    Note that more aggressive data augmentation can be used to
    improve the performances. See the following article for more
    information:

        Wightman R., Touvron H., Jégou H.
        ResNet strikes back: An improved training procedure in timm.
        http://arxiv.org/pdf/2110.00476

    Args:
        train_path (``pathlib.Path`` or str or ``None``): Specifies
            the path to the training dataset. If ``None``, the
            training dataset is set to ``None``.
        eval_path (``pathlib.Path`` or str or ``None``): Specifies
            the path to the evaluation dataset. If ``None``, the
            evaluation dataset is set to ``None``.
        input_size (int, optional): Specifies the input size for the
            image. Default: ``224``

    Returns:
        (``Dataset`` or ``None``, ``Dataset`` or ``None``):
            The training and evaluation datasets. The datasets use the
            original image data augmentation (V1).
    """
    check_torchvision()
    normalize = transforms.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
    train_dataset, eval_dataset = None, None
    if train_path:
        logger.info(f"Initialize the training dataset from {train_path}")
        train_dataset = ImageFolderDataset(
            train_path,
            transforms.Compose(
                [
                    transforms.RandomResizedCrop(input_size),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    normalize,
                ]
            ),
        )
        logger.info(f"number of training examples: {len(train_dataset):,}")

    if eval_path:
        logger.info(f"Initialize the evaluation dataset from {eval_path}")
        eval_dataset = ImageFolderDataset(
            eval_path,
            transforms.Compose(
                [
                    # Maintain same ratio w.r.t. 256 -> 224 images
                    transforms.Resize(int((256 / 224) * input_size)),
                    transforms.CenterCrop(input_size),
                    transforms.ToTensor(),
                    normalize,
                ]
            ),
        )
        logger.info(f"number of evaluation examples: {len(eval_dataset):,}")
    return train_dataset, eval_dataset



r"""This module contains some tools to analyze the parameters of a
``torch.nn.Module``."""

__all__ = [
    "analyze_module_architecture",
    "analyze_model_architecture",
    "analyze_model_architecture",
]

import logging
from typing import Optional

from torch.nn import Module

from gravitorch.engines.base import BaseEngine
from gravitorch.models.utils.summary import ModelSummary
from gravitorch.nn.utils.module_helpers import (
    get_module_devices,
    num_learnable_parameters,
    num_parameters,
)
from gravitorch.utils.exp_trackers import EpochStep

logger = logging.getLogger(__name__)


def analyze_module_architecture(
    module: Module, engine: Optional[BaseEngine] = None, prefix: str = ""
) -> None:
    r"""Analyzes the architecture of a ``torch.nn.Module`` object.

    This function does nothing if the module is not
    ``torch.nn.Module``.

    Args:
        module (``torch.nn.Module``): Specifies the module to analyze.
        engine (``BaseEngine`` or ``None``, optional): Specifies the
            engine to use to log the info. If ``None``, no metric will
            be logged about the model. Default: ``None``
        prefix (str, optional): Specifies the prefix used to log the
            number of parameters. Default: ``''``
    """
    if not isinstance(module, Module):
        return
    logger.info(f"Architecture analysis of the top layers\n{ModelSummary(module, mode='top')}")
    logger.info(f"Architecture analysis of all the layers\n{ModelSummary(module, mode='full')}")
    logger.info(f"devices: {get_module_devices(module)}")
    if engine:
        engine.log_metrics(
            {
                f"{prefix}num_parameters": num_parameters(module),
                f"{prefix}num_learnable_parameters": num_learnable_parameters(module),
            },
            step=EpochStep(engine.epoch),
        )


def analyze_model_architecture(model: Module, engine: Optional[BaseEngine] = None) -> None:
    r"""Analyzes the architecture of a model.

    Args:
        model (``torch.nn.Module``): Specifies the model to log info.
        engine (``BaseEngine`` or ``None``, optional): Specifies the
            engine to use to log the info. If ``None``, no metric
            will be logged about the module. Default: ``None``
    """
    analyze_module_architecture(module=model, engine=engine, prefix="model.")


def analyze_model_network_architecture(model: Module, engine: Optional[BaseEngine] = None) -> None:
    r"""Analyzes the network architecture of a model.

    This function assumes the model has a  ``network`` attribute.
    This function only analyzes the ``model.network`` attribute,
    and it does nothing if ``model.network`` is not a
    ``torch.nn.Module``.

    Args:
        model (``torch.nn.Module``): Specifies the model to analyze.
        engine (``BaseEngine`` or ``None``, optional): Specifies the
            engine to use to log the info. If ``None``, no metric
            will be logged about the module. Default: ``None``
    """
    if hasattr(model, "network"):
        analyze_module_architecture(module=model.network, engine=engine, prefix="model.network.")
    else:
        logger.info("The module does not have a network attribute")



r"""This module implements datasets that stores all the examples in memory."""

__all__ = ["InMemoryDataset", "FileToInMemoryDataset"]

import logging
from collections.abc import Sequence
from pathlib import Path
from typing import TypeVar, Union

import torch
from torch.utils.data import Dataset

from gravitorch.data.datasets.utils import log_box_dataset_class
from gravitorch.utils.io import load_json, load_pickle
from gravitorch.utils.path import sanitize_path

logger = logging.getLogger(__name__)

T = TypeVar("T")


class InMemoryDataset(Dataset[T]):
    r"""Implements a dataset that stores all the examples in-memory.

    You can use this dataset only if all the examples can fit
    in-memory.

    Args:
        examples: Specifies the examples of the dataset.
    """

    def __init__(self, examples: Sequence[T]):
        log_box_dataset_class(self)
        self._examples = tuple(examples)

    def __len__(self) -> int:
        return len(self._examples)

    def __getitem__(self, item: int) -> T:
        return self._examples[item]

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(num_examples={len(self):,})"


class FileToInMemoryDataset(Dataset[T]):
    r"""Implements a dataset that loads examples from a file and store them in
    memory.

    The data in the files should already be preprocessed and organized
    by examples. The file should store a tuple (or list) of examples.
    You can use this dataset only if all the examples can fit
    in-memory.

    This dataset supports the following file formats:

        - PyTorch (file created by ``torch.save``): the extension of
            the file has to be ``.pt``
        - pickle: the extension of the file has to be ``.pkl``
        - json: the extension of the file has to be ``.json``

    The extension of the file is used to find the loader to use.

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the file
            to load.
    """

    def __init__(self, path: Union[Path, str]):
        log_box_dataset_class(self)
        self._path = sanitize_path(path)
        logger.info(f"Loading data from: {self._path}")
        self._examples = _load_examples(self._path)

    def __len__(self) -> int:
        return len(self._examples)

    def __getitem__(self, item: int) -> T:
        return self._examples[item]

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(num_examples={len(self):,}, path={self._path})"


def _load_examples(path: Path) -> tuple[T, ...]:
    r"""Loads the examples from a file.

    This function supports the following file formats:

        - PyTorch (file created by ``torch.save``): the extension of
            the file has to be ``.pt``
        - pickle: the extension of the file has to be ``.pkl``
        - json: the extension of the file has to be ``.json``

    Args:
        path (``pathlib.Path`` or str): Specifies the path to the file
            to load.
    """
    if path.suffix == ".pt":
        return tuple(torch.load(path))
    if path.suffix == ".pkl":
        return tuple(load_pickle(path))
    if path.suffix == ".json":
        return tuple(load_json(path))
    raise ValueError(
        f"Incorrect file extension {path.suffix}. The supported file extensions "
        "are '.pt', '.pkl' and '.json'"
    )



r"""This package contains the implementation of some data loader creators."""

__all__ = [
    "AutoDataLoaderCreator",
    "BaseDataLoaderCreator",
    "DistributedDataLoaderCreator",
    "VanillaDataLoaderCreator",
    "setup_data_loader_creator",
]

from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.creators.dataloader.pytorch import (
    AutoDataLoaderCreator,
    DistributedDataLoaderCreator,
    VanillaDataLoaderCreator,
)
from gravitorch.creators.dataloader.utils import setup_data_loader_creator



__all__ = [
    "BaseOptimizerCreator",
    "NoOptimizerCreator",
    "VanillaOptimizerCreator",
    "ZeroRedundancyOptimizerCreator",
    "setup_optimizer_creator",
]

from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.creators.optimizer.noo import NoOptimizerCreator
from gravitorch.creators.optimizer.utils import setup_optimizer_creator
from gravitorch.creators.optimizer.vanilla import VanillaOptimizerCreator
from gravitorch.creators.optimizer.zero import ZeroRedundancyOptimizerCreator



__all__ = ["ModelArchitectureAnalyzer", "ModelNetworkArchitectureAnalyzer"]

from typing import Union

from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.handlers.base import BaseHandler
from gravitorch.handlers.utils import to_events
from gravitorch.models.utils import (
    analyze_model_architecture,
    analyze_model_network_architecture,
)
from gravitorch.utils.events import VanillaEventHandler


class ModelArchitectureAnalyzer(BaseHandler):
    r"""Implements a handler to analyze a model architecture.

    Args:
        events (str or tuple or list): Specifies the event(s) when to
            analyze the model architecture. It is usually a good idea
            to log model information at the beginning of the training.
            Default: ``('started',)``
    """

    def __init__(
        self,
        events: Union[str, tuple[str, ...], list[str]] = (EngineEvents.STARTED,),
    ):
        self._events = to_events(events)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(events={self._events})"

    def attach(self, engine: BaseEngine) -> None:
        for event in self._events:
            engine.add_event_handler(
                event,
                VanillaEventHandler(self.analyze, handler_kwargs={"engine": engine}),
            )

    def analyze(self, engine: BaseEngine) -> None:
        r"""Analyzes the model architecture.

        Args:
            engine (``BaseEngine``): Specifies the engine with the
                model to analyze.
        """
        analyze_model_architecture(model=engine.model, engine=engine)


class ModelNetworkArchitectureAnalyzer(ModelArchitectureAnalyzer):
    r"""Implements a handler to analyze the network architecture of a model.

    This handler assumes the model has an attribute ``network``.
    """

    def analyze(self, engine: BaseEngine) -> None:
        r"""Analyzes the network architecture of a model.

        Args:
            engine (``BaseEngine``): Specifies the engine with the
                model to analyze.
        """
        analyze_model_network_architecture(model=engine.model, engine=engine)



__all__ = ["AdvancedCoreCreator"]

from typing import Optional, Union

from torch.nn import Module
from torch.optim import Optimizer

from gravitorch.creators.core.base import BaseCoreCreator
from gravitorch.creators.datasource.base import (
    BaseDataSourceCreator,
    setup_data_source_creator,
)
from gravitorch.creators.lr_scheduler import (
    BaseLRSchedulerCreator,
    setup_lr_scheduler_creator,
)
from gravitorch.creators.model import BaseModelCreator, setup_model_creator
from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.creators.optimizer.utils import setup_optimizer_creator
from gravitorch.datasources.base import BaseDataSource
from gravitorch.engines.base import BaseEngine
from gravitorch.lr_schedulers.base import LRSchedulerType
from gravitorch.utils.format import str_add_indent


class AdvancedCoreCreator(BaseCoreCreator):
    r"""Implements an advanced core engine moules creator.

    Args:
        data_source_creator (``BaseDataSourceCreator`` or dict):
            Specifies the data source creator or its configuration.
        model_creator (``BaseModelCreator`` or dict): Specifies the
            model creator or its configuration.
        optimizer_creator (``BaseOptimizerCreator`` or dict or
            ``None`): Specifies the optimizer creator or its
            configuration. Default: ``None``
        lr_scheduler_creator (``BaseLRSchedulerCreator`` or dict or
            ``None`): Specifies the LR scheduler creator or its
            configuration. Default: ``None``
    """

    def __init__(
        self,
        data_source_creator: Union[BaseDataSourceCreator, dict],
        model_creator: Union[BaseModelCreator, dict],
        optimizer_creator: Union[BaseOptimizerCreator, dict, None] = None,
        lr_scheduler_creator: Union[BaseLRSchedulerCreator, dict, None] = None,
    ):
        self._data_source_creator = setup_data_source_creator(data_source_creator)
        self._model_creator = setup_model_creator(model_creator)
        self._optimizer_creator = setup_optimizer_creator(optimizer_creator)
        self._lr_scheduler_creator = setup_lr_scheduler_creator(lr_scheduler_creator)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  data_source_creator={str_add_indent(self._data_source_creator)},\n"
            f"  model_creator={str_add_indent(self._model_creator)},\n"
            f"  optimizer_creator={str_add_indent(self._optimizer_creator)},\n"
            f"  lr_scheduler_creator={str_add_indent(self._lr_scheduler_creator)},\n"
            ")"
        )

    def create(
        self, engine: BaseEngine
    ) -> tuple[BaseDataSource, Module, Optional[Optimizer], Optional[LRSchedulerType]]:
        data_source = self._data_source_creator.create(engine=engine)
        model = self._model_creator.create(engine=engine)
        optimizer = self._optimizer_creator.create(engine=engine, model=model)
        lr_scheduler = self._lr_scheduler_creator.create(engine=engine, optimizer=optimizer)
        return data_source, model, optimizer, lr_scheduler



__all__ = [
    "BaseAlphaActivation",
    "ExpSin",
    "Gaussian",
    "Laplacian",
    "MultiQuadratic",
    "Quadratic",
    "Sin",
]

import torch
from torch import Tensor
from torch.nn import Module, Parameter


class BaseAlphaActivation(Module):
    r"""Defines a base class to implement an activation layer with a learnable
    parameter ``alpha``.

    When called without arguments, the activation layer uses a single
    parameter ``alpha`` across all input channels. If called with a
    first argument, a separate ``alpha`` is used for each input
    channel.

    Args:
        num_parameters (int, optional): Specifies the number of
            learnable parameters. Although it takes an integer as
            input, there is only two values are legitimate: ``1``,
            or the number of channels at input. Default: ``1``
        init (float, optional): Specifies the initial value of the
            learnable parameter(s). Default: ``1.0``
        learnable (bool, optional): If ``True``, the parameters are
            learnt during the training, otherwise they are fixed.
            Default: ``True``
    """

    def __init__(self, num_parameters: int = 1, init: float = 1.0, learnable: bool = True):
        super().__init__()
        self.alpha = Parameter(
            torch.full((num_parameters,), init, dtype=torch.float), requires_grad=learnable
        )

    def extra_repr(self) -> str:
        return f"num_parameters={self.alpha.numel()}, learnable={self.alpha.requires_grad}"


class Gaussian(BaseAlphaActivation):
    r"""Implements the Gaussian activation layer.

    Formula: ``exp(-0.5 * x^2 / alpha^2)``

    This activation layer was proposed in the following paper:

        Beyond Periodicity: Towards a Unifying Framework for Activations in Coordinate-MLPs.
        Ramasinghe S., Lucey S.
        ECCV 2022. (http://arxiv.org/pdf/2111.15135)
    """

    def forward(self, tensor: Tensor) -> Tensor:
        return tensor.pow(2).mul(-0.5).div(self.alpha.pow(2)).exp()


class Laplacian(BaseAlphaActivation):
    r"""Implements the Laplacian activation layer.

    Formula: ``exp(-|x| / alpha)``

    This activation layer was proposed in the following paper:

        Beyond Periodicity: Towards a Unifying Framework for Activations in Coordinate-MLPs.
        Ramasinghe S., Lucey S.
        ECCV 2022. (http://arxiv.org/pdf/2111.15135)
    """

    def forward(self, tensor: Tensor) -> Tensor:
        return tensor.abs().mul(-1).div(self.alpha).exp()


class ExpSin(BaseAlphaActivation):
    r"""Implements the ExpSin activation layer.

    Formula: ``exp(-sin(alpha * x))``

    This activation layer was proposed in the following paper:

        Beyond Periodicity: Towards a Unifying Framework for Activations in Coordinate-MLPs.
        Ramasinghe S., Lucey S.
        ECCV 2022. (http://arxiv.org/pdf/2111.15135)
    """

    def forward(self, tensor: Tensor) -> Tensor:
        return tensor.mul(self.alpha).sin().exp()


class Sin(BaseAlphaActivation):
    r"""Implements the sine activation layer.

    Formula: ``sin(alpha * x)``
    """

    def forward(self, tensor: Tensor) -> Tensor:
        return tensor.mul(self.alpha).sin()


class Quadratic(BaseAlphaActivation):
    r"""Implements the Quadratic activation layer.

    Formula: ``1 / (1 + (alpha * x)^2)``

    This activation layer was proposed in the following paper:

        Beyond Periodicity: Towards a Unifying Framework for Activations in Coordinate-MLPs.
        Ramasinghe S., Lucey S.
        ECCV 2022. (http://arxiv.org/pdf/2111.15135)
    """

    def forward(self, tensor: Tensor) -> Tensor:
        return 1.0 / tensor.mul(self.alpha).pow(2).add(1)


class MultiQuadratic(BaseAlphaActivation):
    r"""Implements the Multi Quadratic activation layer.

    Formula: ``1 / sqrt(1 + (alpha * x)^2)``

    This activation layer was proposed in the following paper:

        Beyond Periodicity: Towards a Unifying Framework for Activations in Coordinate-MLPs.
        Ramasinghe S., Lucey S.
        ECCV 2022. (http://arxiv.org/pdf/2111.15135)
    """

    def forward(self, tensor: Tensor) -> Tensor:
        return 1.0 / tensor.mul(self.alpha).pow(2).add(1).sqrt()



r"""This module defines some primitives for distributed communication."""

__all__ = [
    "Backend",
    "UnknownBackendError",
    "all_gather",
    "all_reduce",
    "auto_backend",
    "available_backends",
    "backend",
    "barrier",
    "broadcast",
    "device",
    "distributed_context",
    "finalize",
    "get_local_rank",
    "get_nnodes",
    "get_node_rank",
    "get_nproc_per_node",
    "get_rank",
    "get_world_size",
    "gloocontext",
    "hostname",
    "initialize",
    "is_distributed",
    "is_main_process",
    "model_name",
    "ncclcontext",
    "resolve_backend",
    "set_local_rank",
    "show_config",
]

import logging
from collections.abc import Generator
from contextlib import contextmanager, nullcontext
from typing import Optional

import torch
from ignite.distributed import utils

from gravitorch.distributed.utils import is_distributed_ready, show_distributed_env_vars

logger = logging.getLogger(__name__)


class Backend:
    r"""Defines the name of the distributed backends currently supported."""
    NCCL = "nccl"
    GLOO = "gloo"


# Do not use ignite directly because it will give more freedom if we want to change one day.
# Only this file should call directly the PyTorch Ignite functions.
all_gather = utils.all_gather
all_reduce = utils.all_reduce
available_backends = utils.available_backends
backend = utils.backend
barrier = utils.barrier
broadcast = utils.broadcast
device = utils.device
finalize = utils.finalize
get_local_rank = utils.get_local_rank
get_nnodes = utils.get_nnodes
get_node_rank = utils.get_node_rank
get_nproc_per_node = utils.get_nproc_per_node
get_rank = utils.get_rank
get_world_size = utils.get_world_size
hostname = utils.hostname
initialize = utils.initialize
model_name = utils.model_name
set_local_rank = utils.set_local_rank
show_config = utils.show_config


class UnknownBackendError(Exception):
    r"""This exception is raised when you try to use an unknown backend."""


def is_main_process() -> bool:
    r"""Indicates if this process is the main process.

    By definition, the main process is the process with the global
    rank 0.

    Returns:
        bool: ``True`` if it is the main process, otherwise ``False``.
    """
    return get_rank() == 0


def is_distributed():
    r"""Indicates if the current process is part of a distributed group.

    Returns:
        bool: ``True`` if the current process is part of a distributed
            group, otherwise ``False``.
    """
    return get_world_size() > 1


@contextmanager
def distributed_context(backend: str) -> Generator[None, None, None]:
    r"""Context manager to initialize the distributed context for a given
    backend.

    Args:
        backend (str): Specifies the distributed backend to use.
            You can find more information on the distributed backends
            at https://pytorch.org/docs/stable/distributed.html#backends

    Example usage

    .. code-block:: python

        >>> import torch
        >>> from gravitorch import distributed as dist
        >>> with dist.distributed_context(backend='gloo'):
        ...     print(dist.backend())
        ...     x = torch.ones(2, 3, device=dist.device())
        ...     dist.all_reduce(x, op="SUM")
        ...     print(x)
        gloo
        tensor([[1., 1., 1.],
                [1., 1., 1.]])
    """
    show_distributed_env_vars()
    if backend not in available_backends():
        raise UnknownBackendError(
            f"Unknown backend '{backend}'. Available backends: {available_backends()}"
        )

    initialize(backend, init_method="env://")

    try:
        # Distributed processes synchronization is needed here to
        # prevent a possible timeout after calling init_process_group.
        # See: https://github.com/facebookresearch/maskrcnn-benchmark/issues/172
        barrier()
        yield
    finally:
        logger.info("Destroying the distributed process...")
        finalize()
        logger.info("Distributed process destroyed")


def auto_backend() -> Optional[str]:
    r"""Finds the best distributed backend for the current environment.

    The rules to find the best distributed backend are:

        - If the NCCL backend and a GPU are available, the best
            distributed backend is NCCL
        - If the GLOO backend is available, the best distributed
            backend is GLOO
        - Otherwise, ``None`` is returned because there is no
            best distributed backend

    Returns:
        str or ``None``: The name of the best distributed backend.

    Example usage:

    .. code-block:: python

        >>> from gravitorch import distributed as dist
        >>> dist.auto_backend()
        'gloo'
    """
    if torch.cuda.is_available() and Backend.NCCL in available_backends():
        return Backend.NCCL
    if Backend.GLOO in available_backends():
        return Backend.GLOO
    return None


def resolve_backend(backend: Optional[str]) -> Optional[str]:
    r"""Resolves the distributed backend if ``'auto'``.

    Args:
        backend (str or ``None``): Specifies the distributed
            backend. If ``'auto'``, this function will find the best
            option for the distributed backend according to the
            context and some rules.

    Returns:
        str or ``None``: The distributed backend or ``None`` if it
            should not use a distributed backend.

    Example usage:

    .. code-block:: python

        >>> from gravitorch import distributed as dist
        >>> dist.resolve_backend("auto")
        gloo
    """
    if backend is None:
        return None
    if backend == "auto":
        if is_distributed_ready():
            return auto_backend()
        # Set to ``None`` because the process does not seem ready
        # to be configured for a distributed experiment.
        return None
    if backend not in available_backends():
        raise UnknownBackendError(
            f"Unknown distributed backend '{backend}'. Available backends: {available_backends()}"
        )
    return backend


@contextmanager
def gloocontext() -> Generator[None, None, None]:
    r"""Context manager to initialize a GLOO distributed context.

    Example usage:

    .. code-block:: python

        >>> from gravitorch.distributed import backend, gloocontext
        >>> with gloocontext():
        ...     print(backend())
        gloo
    """
    if Backend.GLOO not in available_backends():
        raise RuntimeError(
            f"GLOO backend is not available. Available backends: {available_backends()}"
        )
    with distributed_context(Backend.GLOO):
        yield


@contextmanager
def ncclcontext() -> Generator[None, None, None]:
    r"""Context manager to initialize the NCCL distributed context.

    Raises:
        RuntimeError if CUDA is not available

    Example usage:

    .. code-block:: python

        >>> from gravitorch.distributed import backend, ncclcontext
        >>> with ncclcontext():
        ...     print(backend())
        nccl
    """
    if Backend.NCCL not in available_backends():
        raise RuntimeError(
            f"NCCL backend is not available. Available backends: {available_backends()}"
        )
    if not torch.cuda.is_available():
        raise RuntimeError("NCCL backend requires CUDA capable devices but CUDA is not available")
    with distributed_context(Backend.NCCL):
        with torch.cuda.device(get_local_rank()):
            yield


BACKEND_TO_CONTEXT = {
    Backend.NCCL: ncclcontext(),
    Backend.GLOO: gloocontext(),
    None: nullcontext(),
}



r"""This module defines some accuracy metrics."""

__all__ = ["BinaryAccuracy", "CategoricalAccuracy", "TopKAccuracy"]

import logging
from collections.abc import Sequence
from typing import Optional, Union

from objectory import OBJECT_TARGET
from torch import Tensor
from torch.nn import Identity

from gravitorch.engines.base import BaseEngine
from gravitorch.models.metrics.base_epoch import BaseEpochMetric, BaseStateEpochMetric
from gravitorch.models.metrics.state import AccuracyState, BaseState, setup_state
from gravitorch.nn import ToBinaryLabel, ToCategoricalLabel
from gravitorch.utils.exp_trackers import EpochStep
from gravitorch.utils.format import str_add_indent, to_torch_mapping_str

logger = logging.getLogger(__name__)


class BinaryAccuracy(BaseStateEpochMetric):
    r"""Implements the binary accuracy metric.

    Args:
        mode (str): Specifies the mode.
        threshold (float or ``None``, optional): Specifies a threshold
            value to generate the predicted labels from the
            predictions. If ``None``, the predictions are interpreted
            as the predicted labels. Default: ``None``
        name (str, optional): Specifies the name used to log the
            metric. Default: ``'bin_acc'``
        state (``BaseState`` or dict or ``None``, optional): Specifies
            the metric state or its configuration. If ``None``,
            ``AccuracyState`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        mode: str,
        threshold: Optional[float] = None,
        name: str = "bin_acc",
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or AccuracyState())
        self.prediction_transform = Identity() if threshold is None else ToBinaryLabel(threshold)

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the binary accuracy metric given a mini-batch of examples.

        Args:
            prediction (``torch.Tensor`` of type float and shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)``
                and type bool or long or float):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)``
                and type bool or long or float):
                Specifies the targets. The values have to be ``0`` or
                ``1``.
        """
        prediction = self.prediction_transform(prediction)
        self._state.update(prediction.eq(target.view_as(prediction)))


class CategoricalAccuracy(BaseStateEpochMetric):
    r"""Implements a categorical accuracy metric.

    Args:
        mode (str): Specifies the mode.
        name (str, optional): Specifies the name used to log the
            metric. Default: ``'cat_acc'``
        state (``BaseState`` or dict or ``None``, optional): Specifies
            the metric state or its configuration. If ``None``,
            ``AccuracyState`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        mode: str,
        name: str = "cat_acc",
        state: Union[BaseState, dict, None] = None,
    ):
        super().__init__(mode=mode, name=name, state=state or AccuracyState())
        self.prediction_transform = ToCategoricalLabel()

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the accuracy metric given a mini-batch of examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn, num_classes)`` and type float):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)`` and
                type long or float): Specifies the categorical
                targets. The values have to be in
                ``{0, 1, ..., num_classes-1}``.
        """
        prediction = self.prediction_transform(prediction)
        self._state.update(prediction.eq(target.view_as(prediction)))


class TopKAccuracy(BaseEpochMetric):
    r"""Implements the accuracy at k metric a.k.a. top-k accuracy.

    Args:
        mode (str): Specifies the mode.
        topk (list or tuple, optional): Specifies the k values used to
            evaluate the top-k accuracy metric. Default: ``(1, 5)``
        name (str, optional): Specifies the name used to log the
            metric. Default: ``'accuracy'``
    """

    def __init__(
        self,
        mode: str,
        topk: Sequence[int] = (1, 5),
        name: str = "acc_top",
        state_config: Optional[dict] = None,
    ):
        super().__init__(mode, name)
        self._topk = topk if isinstance(topk, tuple) else tuple(topk)
        self._maxk = max(self._topk)

        self._states: dict[int, BaseState] = {
            tol: setup_state(
                state_config or {OBJECT_TARGET: "gravitorch.models.metrics.state.AccuracyState"}
            )
            for tol in self._topk
        }

    def extra_repr(self) -> str:
        return (
            f"mode={self._mode},\nname={self._name},\ntopk={self._topk},\n"
            f"states:\n  {str_add_indent(to_torch_mapping_str(self._states))}"
        )

    @property
    def topk(self) -> tuple[int, ...]:
        return self._topk

    def attach(self, engine: BaseEngine) -> None:
        r"""Attaches current metric to the provided engine.

        This method can be used to:

            - add event handler to the engine
            - set up history trackers

        Args:
            engine (``BaseEngine``): Specifies the engine.
        """
        super().attach(engine)
        for k, state in self._states.items():
            for history in state.get_histories(f"{self._metric_name}_{k}_"):
                engine.add_history(history)

    def forward(self, prediction: Tensor, target: Tensor) -> None:
        r"""Updates the accuracy metric given a mini-batch of examples.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn, num_classes)`` and type float):
                Specifies the predictions.
            target (``torch.Tensor`` of shape
                ``(d0, d1, ..., dn)`` or ``(d0, d1, ..., dn, 1)``
                and type long or float): Specifies the targets.
                The values have to be in
                ``{0, 1, ..., num_classes-1}``.
        """
        _, pred = prediction.topk(self._maxk, -1, True, True)
        correct = pred.eq(target.view(*pred.shape[:-1], 1).expand_as(pred)).float()
        for k, state in self._states.items():
            state.update(correct[..., :k].sum(dim=-1))

    def reset(self) -> None:
        r"""Resets the metric."""
        for state in self._states.values():
            state.reset()

    def value(self, engine: Optional[BaseEngine] = None) -> dict:
        r"""Evaluates the metric and log the results given all the examples
        previously seen.

        Args:
            engine (``BaseEngine``, optional): Specifies the engine.
                This argument is required to log the results.
                Default: ``None``.

        Returns:
             dict: The results of the metric
        """
        results = {}
        for k, state in self._states.items():
            results.update(state.value(f"{self._metric_name}_{k}_"))
        if engine:
            engine.log_metrics(results, EpochStep(engine.epoch))
        return results



r"""This module contains implementation to use PyTorch criterion on sequences
with a warm-up approach."""

__all__ = ["WarmupSequenceLoss"]

from typing import Union

from torch import Tensor
from torch.nn import Module

from gravitorch.nn import setup_nn_module


class WarmupSequenceLoss(Module):
    r"""Implements a wrapper to adapt PyTorch loss on sequences with a warm-up
    approach.

    The loss is computed only on the time steps after the warm-up to
    give the model time to warm-up its internal state before making a
    prediction. If ``warmup=5``, the loss is not computed on the first
    five time steps. This loss function assumes that all the sequences
    have the same length.

    This wrapper assumes the loss function has the following two
    inputs:

        - a tensor of prediction of shape
            ``(sequence_length, batch_size, *)`` if
            ``batch_size=False`` or
            ``(batch_size, sequence_length, *)`` otherwise.
        - a tensor of target of shape
            ``(sequence_length, batch_size, *)`` if
            ``batch_size=False`` or
            ``(batch_size, sequence_length, *)`` otherwise.

    Internally, this loss function converts the batch of sequences
    as a batch of examples i.e. the sequence and batch dimensions are
    merged together. Most of the PyTorch criteria work on a batch
    examples i.e. tensors of dimension ``(batch_size, *)`` where ``*``
    means, any number of additional dimensions.

    Note: this is an experimental loss function.

    Args:
        criterion (``torch.nn.Module`` or dict): Specifies the loss
            function/criterion or its configuration.
        warmup (int, optional): Specifies the number of warmup steps.
            Note that this value should be compatible with
            input dimension. Default: ``0``
        batch_first (bool, optional): Indicates if the first dimension
            is the batch or the sequence. If ``True``, the input
            sequence should have the shape
            ``(batch_size, sequence_length, *)``, otherwise
            ``(sequence_length, batch_size, *)``. Default: ``False``
    """

    def __init__(
        self,
        criterion: Union[Module, dict],
        warmup: int = 0,
        batch_first: bool = False,
    ):
        super().__init__()
        self.criterion = setup_nn_module(criterion)
        self._warmup = int(warmup)
        self._batch_first = bool(batch_first)

    def forward(self, prediction: Tensor, target: Tensor) -> Tensor:
        r"""Computes the criterion function given the network output and the
        batch.

        Args:
            prediction (``torch.Tensor`` of shape
                ``(sequence_length, batch_size, *)`` if
                ``batch_size=False`` or
                ``(batch_size, sequence_length, *)`` otherwise):
                Specifies the predictions. The tensor type depends on
                the loss function used.
            target (``torch.Tensor`` of shape
                ``(sequence_length, batch_size, *)`` if
                ``batch_size=False`` or
                ``(batch_size, sequence_length, *)`` otherwise):
                Specifies the targets. The tensor type depends on the
                loss function used.

        Returns:
            ``torch.Tensor``: The computed loss value. The shape of
                the tensor depends on the loss function used.
        """
        # Remove the warm-up steps from the prediction and target
        if self._batch_first:
            prediction = prediction[:, self._warmup :]
            target = target[:, self._warmup :]
        else:
            prediction = prediction[self._warmup :, :]
            target = target[self._warmup :, :]
        return self.criterion(prediction.flatten(0, 1), target.flatten(0, 1))



__all__ = ["DistributedContext"]

import logging
from types import TracebackType
from typing import Optional

from gravitorch.distributed.comm import BACKEND_TO_CONTEXT, resolve_backend
from gravitorch.distributed.utils import show_distributed_context_info
from gravitorch.rsrc.base import BaseResource

logger = logging.getLogger(__name__)


class DistributedContext(BaseResource):
    r"""Implements a context manager to initialize the distributed backend.

    Args:
        backend (str or ``None``, optional): Specifies the distributed
            backend. If ``'auto'``, this function will find the best
            option for the distributed backend according to the
            context and some rules. Default: ``'auto'``
        log_info (bool, optional): If ``True``, information about the
            resource is logged after the resource is initialized.
            Default: ``False``
    """

    def __init__(self, backend: Optional[str] = "auto", log_info: bool = False):
        self._backend = resolve_backend(backend)
        self._context = BACKEND_TO_CONTEXT[self._backend]

        self._log_info = bool(log_info)

    def __enter__(self) -> "DistributedContext":
        logger.info(f"Initialing `{self._backend}` distributed context...")
        self._context.__enter__()
        if self._log_info:
            show_distributed_context_info()
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        logger.info(f"Exiting `{self._backend}` distributed context...")
        self._context.__exit__(exc_type, exc_val, exc_tb)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(backend={self._backend}, log_info={self._log_info})"



__all__ = [
    "asinh_mse_loss",
    "log_cosh_loss",
    "msle_loss",
    "relative_mse_loss",
    "relative_smooth_l1_loss",
    "symlog_mse_loss",
    "symmetric_relative_smooth_l1_loss",
]

from torch import Tensor
from torch.nn import functional as F

from gravitorch.nn.functional.loss_helpers import basic_loss_reduction
from gravitorch.utils.tensor.math_ops import symlog


def msle_loss(prediction: Tensor, target: Tensor, reduction: str = "mean") -> Tensor:
    r"""Computes the mean squared logarithmic error (MSLE).

    This loss is best to use when targets having exponential growth,
    such as population counts, average sales of a commodity over a
    span of years etc. Note that this loss penalizes an
    under-predicted estimate greater than an over-predicted estimate.

    Note: this loss only works with positive values (0 included).

    Args:
        prediction (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the predictions.
        target (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the target values.
        reduction (string, optional): Specifies the reduction to apply
            to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``:
            the sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``

    Returns:
        ``torch.Tensor`` of type float: The mean squared logarithmic
            error. The shape of the tensor depends on the reduction
            strategy.
    """
    return F.mse_loss(prediction.log1p(), target.log1p(), reduction=reduction)


def asinh_mse_loss(prediction: Tensor, target: Tensor, reduction: str = "mean") -> Tensor:
    r"""Computes the mean squared error (MSE) on the arcsinh transformed
    prediction and target.

    It is a generalization of mean squared logarithmic error (MSLE)
    that works for real values. The ``arcsinh`` transformation is used
    instead of ``log1p`` because ``arcsinh`` works on negative values.

    Args:
        prediction (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the predictions.
        target (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the target values.
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``

    Returns:
        ``torch.Tensor`` of type float: The mean squared error (MSE)
            on the arcsinh transformed prediction and target. The
            shape of the tensor depends on the reduction strategy.
    """
    return F.mse_loss(prediction.asinh(), target.asinh(), reduction=reduction)


def symlog_mse_loss(prediction: Tensor, target: Tensor, reduction: str = "mean") -> Tensor:
    r"""Computes the mean squared error (MSE) on the symlog transformed
    prediction and target.

    It is a generalization of mean squared logarithmic error (MSLE)
    that works for real values. The ``symlog`` transformation is used
    instead of ``log1p`` because ``symlog`` works on negative values.

    Args:
        prediction (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the predictions.
        target (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the target values.
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``

    Returns:
        ``torch.Tensor`` of type float: The mean squared error (MSE)
            on the symlog transformed prediction and target. The shape
            of the tensor depends on the reduction strategy.
    """
    return F.mse_loss(symlog(prediction), symlog(target), reduction=reduction)


def relative_mse_loss(
    prediction: Tensor, target: Tensor, reduction: str = "mean", eps: float = 1e-8
):
    r"""Computes the relative mean squared error loss.

    Args:
        prediction (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the predictions.
        target (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the target values.
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Returns:
        ``torch.Tensor`` of type float: The relative mean squared error.
            The shape depends on the reduction strategy.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.functional import relative_mse_loss
        >>> loss = relative_mse_loss(torch.randn(3, 5, requires_grad=True), torch.randn(3, 5))
        >>> loss.backward()
    """
    return basic_loss_reduction(
        F.mse_loss(prediction, target, reduction="none").div(target.detach().pow(2).clamp(min=eps)),
        reduction=reduction,
    )


def relative_smooth_l1_loss(
    prediction: Tensor,
    target: Tensor,
    reduction: str = "mean",
    beta: float = 1.0,
    eps: float = 1e-8,
):
    r"""Computes the relative smooth l1 loss.

    Args:
        prediction (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the predictions.
        target (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the target values.
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``
        beta (float, optional): Specifies the threshold at which to
            change between L1 and L2 loss. The value must be
            non-negative. Default: ``1.0``
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Returns:
        ``torch.Tensor`` of type float: The relative smooth l1 loss.
            The shape depends on the reduction strategy.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.functional import relative_smooth_l1_loss
        >>> loss = relative_smooth_l1_loss(torch.randn(3, 5, requires_grad=True), torch.randn(3, 5))
        >>> loss.backward()
    """
    return basic_loss_reduction(
        F.smooth_l1_loss(prediction, target, reduction="none", beta=beta).div(
            target.detach().abs().clamp(min=eps)
        ),
        reduction=reduction,
    )


def symmetric_relative_smooth_l1_loss(
    prediction: Tensor,
    target: Tensor,
    reduction: str = "mean",
    beta: float = 1.0,
    eps: float = 1e-8,
):
    r"""Computes the symmetric relative smooth l1 loss.

    Args:
        prediction (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the predictions.
        target (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the target values.
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``
        beta (float, optional): Specifies the threshold at which to
            change between L1 and L2 loss. The value must be
            non-negative. Default: ``1.0``
        eps (float, optional): Specifies an arbitrary small strictly
            positive number to avoid undefined results when the target
            is zero. Default: ``1e-8``

    Returns:
        ``torch.Tensor`` of type float: The symmetric relative smooth
            l1 loss. The shape depends on the reduction strategy.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.functional import symmetric_relative_smooth_l1_loss
        >>> loss = symmetric_relative_smooth_l1_loss(
        ...     torch.randn(3, 5, requires_grad=True), torch.randn(3, 5)
        ... )
        >>> loss.backward()
    """
    return basic_loss_reduction(
        F.smooth_l1_loss(prediction, target, reduction="none", beta=beta).div(
            target.abs().add(prediction.abs()).detach().mul(0.5).clamp(min=eps)
        ),
        reduction=reduction,
    )


def log_cosh_loss(
    prediction: Tensor,
    target: Tensor,
    reduction: str = "mean",
    scale: float = 1.0,
):
    r"""Computes the logarithm of the hyperbolic cosine of the prediction error.

    Args:
        prediction (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the predictions.
        target (``torch.Tensor`` of type float and shape
            ``(d0, d1, ..., dn)``): Specifies the target values.
        reduction (string, optional): Specifies the reduction to
            apply to the output: ``'none'`` | ``'mean'`` | ``'sum'``.
            ``'none'``: no reduction will be applied, ``'mean'``: the
            sum of the output will be divided by the number of
            elements in the output, ``'sum'``: the output will be
            summed. Default: ``'mean'``
        scale (float, optional): Specifies the scale factor.
            Default: ``1.0``

    Returns:
        ``torch.Tensor`` of type float: the logarithm of the
            hyperbolic cosine of the prediction error.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.nn.functional import log_cosh_loss
        >>> loss = log_cosh_loss(torch.randn(3, 5, requires_grad=True), torch.randn(3, 5))
        >>> loss.backward()
    """
    return basic_loss_reduction(
        target.sub(prediction).div(scale).cosh().log(),
        reduction=reduction,
    )



__all__ = ["OneCacheDataCreator"]

import copy
import logging
from typing import Any, Optional, TypeVar, Union

from gravitorch.data.datacreators import BaseDataCreator, setup_data_creator
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_add_indent

logger = logging.getLogger(__name__)

T = TypeVar("T")


class OneCacheDataCreator(BaseDataCreator[T]):
    r"""Implements a data creator that creates the data only once, then cache
    them and return them.

    Args:
        data_creator (``BaseDataCreator`` or dict): Specifies a data
            creator or its configuration.
        deepcopy (bool, optional): If ``True``, a deepcopy of the data
            is done before to return the data. If ``deepcopy`` is
            explicitly set to ``False``, users should ensure that
            the data pipeline does not contain any in-place operations
            over the data to prevent data inconsistency if the
            ``create`` method is called multiple times.
    """

    def __init__(self, data_creator: Union[BaseDataCreator[T], dict], deepcopy: bool = False):
        self._data_creator = setup_data_creator(data_creator)
        self._deepcopy = bool(deepcopy)
        self._is_cache_created = False
        # This variable is used to cache the data. The type depends on the value returned by
        # the function ``create`` of the data creator object.
        self._cached_data: Any = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  data_creator={str_add_indent(self._data_creator)},\n"
            f"  is_cache_created={self._is_cache_created},\n"
            f"  deepcopy={self._deepcopy},\n"
            ")"
        )

    @property
    def data_creator(self) -> BaseDataCreator[T]:
        r"""``BaseDataCreator``: The data creator."""
        return self._data_creator

    @property
    def deepcopy(self) -> bool:
        r"""bool: Indicates if a deepcopy of the data is done before
        to return the data."""
        return self._deepcopy

    def create(self, engine: Optional[BaseEngine] = None) -> T:
        r"""Creates data.

        Args:
            engine (``BaseEngine`` or ``None``): Specifies an engine.
            Default: ``None``

        Return:
            The created data. The returned data depends on the data
                creator.
        """
        if not self._is_cache_created:
            logger.info("Creating data and caching them...")
            self._cached_data = self._data_creator.create(engine)
            self._is_cache_created = True
        data = self._cached_data
        if self._deepcopy:
            try:
                data = copy.deepcopy(data)
            except TypeError:
                logger.warning(
                    "The data can not be deepcopied. "
                    "Please be aware of in-place modification would affect source data"
                )
        return data



__all__ = ["FixedSizePartitioner"]

from collections.abc import Sequence
from typing import Optional, TypeVar

from gravitorch.data.partitioners.base import BasePartitioner
from gravitorch.engines import BaseEngine
from gravitorch.utils.partitioning import fixed_size_partitions

T = TypeVar("T")


class FixedSizePartitioner(BasePartitioner[T]):
    r"""Implements a partitioner that creates fixed-size partitions.

    Args:
        partition_size (int): Specifies the partition size i.e. the
            number of items in each partition.
        drop_last (bool, optional): If ``True``, it drops the last
            items if the number of items is not evenly divisible by
            ``partition_size``.
    """

    def __init__(self, partition_size: int, drop_last: bool = False):
        self._partition_size = int(partition_size)
        self._drop_last = bool(drop_last)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}("
            f"partition_size={self._partition_size:,}, drop_last={self._drop_last})"
        )

    @property
    def drop_last(self) -> bool:
        r"""bool: Indicates if the last items are dropped
        or not if there are not enough items.
        """
        return self._drop_last

    @property
    def partition_size(self) -> int:
        r"""int: The partition size."""
        return self._partition_size

    def partition(
        self, items: Sequence[T], engine: Optional[BaseEngine] = None
    ) -> list[Sequence[T]]:
        return fixed_size_partitions(
            items=items, partition_size=self._partition_size, drop_last=self._drop_last
        )



r"""This module defines some demo datasets."""

__all__ = ["DemoMultiClassClsDataset"]

import logging

from torch.utils.data import Dataset

from gravitorch import constants as ct
from gravitorch.data.datasets.utils import log_box_dataset_class
from gravitorch.utils.summary import concise_summary

logger = logging.getLogger(__name__)


class DemoMultiClassClsDataset(Dataset):
    r"""Implements a toy multi-class classification dataset in a map- style
    format.

    Args:
        num_examples (int, optional): Specifies the number of
            examples. Default: ``1000``
        num_classes (int, optional): Specifies the number of classes.
            Default: 50
        feature_size (int, optional): Specifies the feature size. The
            feature size has to be greater than the number of classes.
            Default: ``64``
        noise_std (float, optional): Specifies the standard deviation
            of the Gaussian noise. Default: ``0.2``
        random_seed (int, optional): Specifies the random seed used to
            initialize a ``torch.Generator`` object.
            Default: ``10169389905513828140``
    """

    def __init__(
        self,
        num_examples: int = 1000,
        num_classes: int = 50,
        feature_size: int = 64,
        noise_std: float = 0.2,
        random_seed: int = 10169389905513828140,
    ):
        log_box_dataset_class(self)

        logger.info("Initializing the data creator...")
        # Import is here to avoid cyclic import
        from gravitorch.data.datacreators import HypercubeVertexDataCreator

        self._data_creator = HypercubeVertexDataCreator(
            num_examples=num_examples,
            num_classes=num_classes,
            feature_size=feature_size,
            noise_std=noise_std,
            random_seed=random_seed,
        )
        logger.info("Creating synthetic data...")
        data = self._data_creator.create()
        logger.info(f"Created data:\n{concise_summary(data)}")

        self._targets = data[ct.TARGET]
        self._features = data[ct.INPUT]

    def __getitem__(self, item: int):
        return {
            ct.INPUT: self._features[item],
            ct.TARGET: self._targets[item],
            ct.NAME: f"{item}",
        }

    def __len__(self) -> int:
        return self.num_examples

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}("
            f"num_examples={self.num_examples:,}, "
            f"num_classes={self.num_classes:,}, "
            f"feature_size={self.feature_size:,}, "
            f"noise_std={self.noise_std:,}, "
            f"random_seed={self.random_seed})"
        )

    @property
    def num_examples(self) -> int:
        r"""int: The number of examples when the data are created."""
        return self._data_creator.num_examples

    @property
    def num_classes(self) -> int:
        r"""int: The number of classes when the data are created."""
        return self._data_creator.num_classes

    @property
    def feature_size(self) -> int:
        r"""int: The feature size when the data are created."""
        return self._data_creator.feature_size

    @property
    def noise_std(self) -> float:
        r"""float: The standard deviation of the Gaussian noise."""
        return self._data_creator.noise_std

    @property
    def random_seed(self) -> int:
        r"""int: The random seed used to initialize a
        ``torch.Generator`` object."""
        return self._data_creator.random_seed



r"""This package contains the implementation of some data loader collators."""

__all__ = [
    "BaseCollator",
    "DictPackedSequenceCollator",
    "DictPaddedSequenceCollator",
    "PackedSequenceCollator",
    "PaddedSequenceCollator",
    "setup_collator",
]

from gravitorch.data.dataloaders.collators.base import BaseCollator
from gravitorch.data.dataloaders.collators.packed_sequence import (
    DictPackedSequenceCollator,
    PackedSequenceCollator,
)
from gravitorch.data.dataloaders.collators.padded_sequence import (
    DictPaddedSequenceCollator,
    PaddedSequenceCollator,
)
from gravitorch.data.dataloaders.collators.utils import setup_collator



__all__ = [
    "AccuracyState",
    "BaseState",
    "ErrorState",
    "ExtendedAccuracyState",
    "ExtendedErrorState",
    "MeanErrorState",
    "RootMeanErrorState",
    "setup_state",
]

import logging
import math
from abc import ABC, abstractmethod
from typing import Union

from objectory import AbstractFactory
from torch import Tensor

from gravitorch.models.metrics.base import EmptyMetricError
from gravitorch.utils.format import str_scalar, str_target_object
from gravitorch.utils.history import BaseHistory, MaxScalarHistory, MinScalarHistory
from gravitorch.utils.meters import MeanTensorMeter, TensorMeter, TensorMeter2
from gravitorch.utils.tensor import to_tensor

logger = logging.getLogger(__name__)


class BaseState(ABC, metaclass=AbstractFactory):
    r"""Defines a base class to implement a metric state.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics.state import ErrorState
        >>> state = ErrorState()
        >>> state.get_histories("error_")
        (MinScalarHistory(name=error_mean, max_history_size=10, history=()),
         MinScalarHistory(name=error_min, max_history_size=10, history=()),
         MinScalarHistory(name=error_max, max_history_size=10, history=()),
         MinScalarHistory(name=error_sum, max_history_size=10, history=()))
        >>> state.update(torch.arange(6))
        >>> state.value("error_")
        {'error_mean': 2.5,
         'error_min': 0.0,
         'error_max': 5.0,
         'error_sum': 15.0,
         'error_num_predictions': 6}
    """

    @property
    @abstractmethod
    def num_predictions(self) -> int:
        r"""int: The number of predictions."""

    @abstractmethod
    def get_histories(self, prefix: str = "", suffix: str = "") -> tuple[BaseHistory, ...]:
        r"""Gets the history trackers for the metric values associated to the
        current state.

        Args:
            prefix (str, optional): Specifies the key prefix in the
                history tracker names. Default: ``''``
            suffix (str, optional): Specifies the key suffix in the
                history tracker names. Default: ``''``

        Returns:
            tuple: The history trackers.
        """

    @abstractmethod
    def reset(self) -> None:
        r"""Resets the state."""

    @abstractmethod
    def update(self, *args, **kwargs) -> None:
        r"""Updates the metric state.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """

    @abstractmethod
    def value(self, prefix: str = "", suffix: str = "") -> dict[str, Union[int, float]]:
        r"""Computes the metric values given the current state.

        Args:
            prefix (str, optional): Specifies the key prefix in the
                returned dictionary. Default: ``''``
            suffix (str, optional): Specifies the key suffix in the
                returned dictionary. Default: ``''``

        Returns:
            dict: The metric values.
        """


class MeanErrorState(BaseState):
    r"""Implements a metric state to capture the mean error value.

    This state has a constant space complexity.

    Args:
        track_num_predictions (bool, optional): If ``True``, the state
            tracks and returns the number of predictions.
            Default: ``True``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics.state import MeanErrorState
        >>> state = MeanErrorState()
        >>> state.get_histories("error_")
        (MinScalarHistory(name=error_mean, max_history_size=10, history=()),)
        >>> state.update(torch.arange(6))
        >>> state.value("error_")
        {'error_mean': 2.5, 'error_num_predictions': 6}
    """

    def __init__(self, track_num_predictions: bool = True):
        self._meter = MeanTensorMeter()
        self._track_num_predictions = bool(track_num_predictions)

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  (meter) {repr(self._meter)}\n)"

    @property
    def num_predictions(self) -> int:
        return self._meter.count

    def get_histories(self, prefix: str = "", suffix: str = "") -> tuple[BaseHistory, ...]:
        return tuple([MinScalarHistory(name=f"{prefix}mean{suffix}")])

    def reset(self) -> None:
        self._meter.reset()

    def update(self, error: Tensor) -> None:
        r"""Updates the metric state with a new tensor of errors.

        Args:
            error (```torch.Tensor``): A tensor of errors.
        """
        self._meter.update(error.detach())

    def value(self, prefix: str = "", suffix: str = "") -> dict[str, Union[int, float]]:
        meter = self._meter.all_reduce()
        if not meter.count:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        results = {f"{prefix}mean{suffix}": self._meter.mean()}
        if self._track_num_predictions:
            results[f"{prefix}num_predictions{suffix}"] = meter.count

        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")
        return results


class RootMeanErrorState(BaseState):
    r"""Implements a metric state to capture the root mean error value.

    This state has a constant space complexity.

    Args:
        track_num_predictions (bool, optional): If ``True``, the state
            tracks and returns the number of predictions.
            Default: ``True``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics.state import RootMeanErrorState
        >>> state = RootMeanErrorState()
        >>> state.get_histories("error_")
        (MinScalarHistory(name=error_root_mean, max_history_size=10, history=()),)
        >>> state.update(torch.arange(6))
        >>> state.value("error_")
        {'error_root_mean': 1.5811388300841898, 'error_num_predictions': 6}
    """

    def __init__(self, track_num_predictions: bool = True):
        self._meter = MeanTensorMeter()
        self._track_num_predictions = bool(track_num_predictions)

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  (meter) {repr(self._meter)}\n)"

    @property
    def num_predictions(self) -> int:
        return self._meter.count

    def get_histories(self, prefix: str = "", suffix: str = "") -> tuple[BaseHistory, ...]:
        return tuple([MinScalarHistory(name=f"{prefix}root_mean{suffix}")])

    def reset(self) -> None:
        self._meter.reset()

    def update(self, error: Tensor) -> None:
        r"""Updates the metric state with a new tensor of errors.

        Args:
            error (```torch.Tensor``): A tensor of errors.
        """
        self._meter.update(error.detach())

    def value(self, prefix: str = "", suffix: str = "") -> dict[str, Union[int, float]]:
        meter = self._meter.all_reduce()
        if not meter.count:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        results = {f"{prefix}root_mean{suffix}": math.sqrt(self._meter.mean())}
        if self._track_num_predictions:
            results[f"{prefix}num_predictions{suffix}"] = meter.count

        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")
        return results


class ErrorState(BaseState):
    r"""Implements a metric state to capture some metrics about the errors.

    This state has a constant space complexity.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics.state import ErrorState
        >>> state = ErrorState()
        >>> state.get_histories("error_")
        (MinScalarHistory(name=error_mean, max_history_size=10, history=()),
         MinScalarHistory(name=error_min, max_history_size=10, history=()),
         MinScalarHistory(name=error_max, max_history_size=10, history=()),
         MinScalarHistory(name=error_sum, max_history_size=10, history=()))
        >>> state.update(torch.arange(6))
        >>> state.value("error_")
        {'error_mean': 2.5,
         'error_min': 0.0,
         'error_max': 5.0,
         'error_sum': 15.0,
         'error_num_predictions': 6}
    """

    def __init__(self):
        self._meter = TensorMeter()

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  (meter) {repr(self._meter)}\n)"

    @property
    def num_predictions(self) -> int:
        return self._meter.count

    def get_histories(self, prefix: str = "", suffix: str = "") -> tuple[BaseHistory, ...]:
        return (
            MinScalarHistory(name=f"{prefix}mean{suffix}"),
            MinScalarHistory(name=f"{prefix}min{suffix}"),
            MinScalarHistory(name=f"{prefix}max{suffix}"),
            MinScalarHistory(name=f"{prefix}sum{suffix}"),
        )

    def reset(self) -> None:
        self._meter.reset()

    def update(self, error: Tensor) -> None:
        r"""Updates the metric state with a new tensor of errors.

        Args:
            error (```torch.Tensor``): A tensor of errors.
        """
        self._meter.update(error.detach())

    def value(self, prefix: str = "", suffix: str = "") -> dict[str, Union[int, float]]:
        meter = self._meter.all_reduce()
        if not meter.count:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        results = {
            f"{prefix}mean{suffix}": self._meter.mean(),
            f"{prefix}min{suffix}": self._meter.min(),
            f"{prefix}max{suffix}": self._meter.max(),
            f"{prefix}sum{suffix}": self._meter.sum(),
            f"{prefix}num_predictions{suffix}": meter.count,
        }
        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")
        return results


class ExtendedErrorState(BaseState):
    r"""Implements a metric state to capture some metrics about the errors.

    This state stores all the error values, so it may not scale to large
    datasets. This state has a linear space complexity.

    Args:
        quantiles (``torch.Tensor`` or list or tuple, optional):
            Specifies the quantile values to evaluate.
            Default: ``tuple()``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics.state import ExtendedErrorState
        >>> state = ExtendedErrorState(quantiles=[0.5, 0.9])
        >>> state.get_histories("error_")
        (MinScalarHistory(name=error_mean, max_history_size=10, history=()),
         MinScalarHistory(name=error_median, max_history_size=10, history=()),
         MinScalarHistory(name=error_min, max_history_size=10, history=()),
         MinScalarHistory(name=error_max, max_history_size=10, history=()),
         MinScalarHistory(name=error_sum, max_history_size=10, history=()),
         MinScalarHistory(name=error_quantile_0.5, max_history_size=10, history=()),
         MinScalarHistory(name=error_quantile_0.9, max_history_size=10, history=()))
        >>> state.update(torch.arange(11))
        >>> state.value("error_")
        {'error_mean': 5.0,
         'error_median': 5,
         'error_min': 0,
         'error_max': 10,
         'error_sum': 55,
         'error_std': 3.316624879837036,
         'error_quantile_0.5': 5.0,
         'error_quantile_0.9': 9.0,
         'error_num_predictions': 11}
    """

    def __init__(self, quantiles: Union[Tensor, tuple[float, ...], list[float]] = tuple()):
        self._meter = TensorMeter2()
        self._quantiles = to_tensor(quantiles)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  (meter) {repr(self._meter)}\n"
            f"  quantiles={self._quantiles}\n)"
        )

    @property
    def num_predictions(self) -> int:
        return self._meter.count

    def get_histories(self, prefix: str = "", suffix: str = "") -> tuple[BaseHistory, ...]:
        trackers = [
            MinScalarHistory(name=f"{prefix}mean{suffix}"),
            MinScalarHistory(name=f"{prefix}median{suffix}"),
            MinScalarHistory(name=f"{prefix}min{suffix}"),
            MinScalarHistory(name=f"{prefix}max{suffix}"),
            MinScalarHistory(name=f"{prefix}sum{suffix}"),
        ]
        for q in self._quantiles:
            trackers.append(
                MinScalarHistory(name=f"{prefix}quantile_{q:g}{suffix}"),
            )
        return tuple(trackers)

    def reset(self) -> None:
        self._meter.reset()

    def update(self, error: Tensor) -> None:
        r"""Updates the metric state with a new tensor of errors.

        Args:
            error (```torch.Tensor``): A tensor of errors.
        """
        self._meter.update(error.detach().cpu())

    def value(self, prefix: str = "", suffix: str = "") -> dict[str, Union[int, float]]:
        meter = self._meter.all_reduce()
        if not meter.count:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        results = {
            f"{prefix}mean{suffix}": self._meter.mean(),
            f"{prefix}median{suffix}": self._meter.median(),
            f"{prefix}min{suffix}": self._meter.min(),
            f"{prefix}max{suffix}": self._meter.max(),
            f"{prefix}sum{suffix}": self._meter.sum(),
            f"{prefix}std{suffix}": self._meter.std(),
        }
        if self._quantiles.numel() > 0:
            values = self._meter.quantile(self._quantiles)
            for q, v in zip(self._quantiles, values):
                results[f"{prefix}quantile_{q:g}{suffix}"] = v.item()
        results[f"{prefix}num_predictions{suffix}"] = meter.count
        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")
        return results


class AccuracyState(BaseState):
    r"""Implements a metric state to compute the accuracy.

    This state has a constant space complexity.

    Args:
        track_num_predictions (bool, optional): If ``True``, the state
            tracks and returns the number of predictions.
            Default: ``True``

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics.state import AccuracyState
        >>> state = AccuracyState()
        >>> state.get_histories()
        (MaxScalarHistory(name=accuracy, max_history_size=10, history=()),)
        >>> state.update(torch.eye(4))
        >>> state.value()
        {'accuracy': 0.25, 'num_predictions': 16}
    """

    def __init__(self, track_num_predictions: bool = True):
        self._meter = MeanTensorMeter()
        self._track_num_predictions = bool(track_num_predictions)

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  (meter) {repr(self._meter)}\n)"

    @property
    def num_predictions(self) -> int:
        return self._meter.count

    def get_histories(self, prefix: str = "", suffix: str = "") -> tuple[BaseHistory, ...]:
        return tuple([MaxScalarHistory(name=f"{prefix}accuracy{suffix}")])

    def reset(self) -> None:
        self._meter.reset()

    def update(self, correct: Tensor) -> None:
        r"""Updates the metric state with a new tensor of errors.

        Args:
            correct (```torch.Tensor``): A tensor that indicates the
                correct predictions. ``1`` indicates a correct
                prediction and ``0`` indicates a bad prediction.
        """
        self._meter.update(correct.detach())

    def value(self, prefix: str = "", suffix: str = "") -> dict[str, Union[int, float]]:
        meter = self._meter.all_reduce()
        if not meter.count:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        results = {f"{prefix}accuracy{suffix}": meter.mean()}
        if self._track_num_predictions:
            results[f"{prefix}num_predictions{suffix}"] = meter.count

        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")
        return results


class ExtendedAccuracyState(BaseState):
    r"""Implements a metric state to compute the accuracy and other metrics.

    This state has a constant space complexity.

    Example usage:

    .. code-block:: python

        >>> import torch
        >>> from gravitorch.models.metrics.state import ExtendedAccuracyState
        >>> state = ExtendedAccuracyState()
        >>> state.get_histories()
        (MaxScalarHistory(name=accuracy, max_history_size=10, history=()),
         MinScalarHistory(name=error, max_history_size=10, history=()),
         MaxScalarHistory(name=num_correct_predictions, max_history_size=10, history=()),
         MinScalarHistory(name=num_incorrect_predictions, max_history_size=10, history=()))
        >>> state.update(torch.eye(4))
        >>> state.value()
        {'accuracy': 0.25,
         'error': 0.75,
         'num_correct_predictions': 4,
         'num_incorrect_predictions': 12,
         'num_predictions': 16}
    """

    def __init__(self):
        self._meter = MeanTensorMeter()

    def __str__(self) -> str:
        return f"{self.__class__.__qualname__}(\n  (meter) {repr(self._meter)}\n)"

    @property
    def num_predictions(self) -> int:
        return self._meter.count

    def get_histories(self, prefix: str = "", suffix: str = "") -> tuple[BaseHistory, ...]:
        return (
            MaxScalarHistory(name=f"{prefix}accuracy{suffix}"),
            MinScalarHistory(name=f"{prefix}error{suffix}"),
            MaxScalarHistory(name=f"{prefix}num_correct_predictions{suffix}"),
            MinScalarHistory(name=f"{prefix}num_incorrect_predictions{suffix}"),
        )

    def reset(self) -> None:
        self._meter.reset()

    def update(self, correct: Tensor) -> None:
        r"""Updates the metric state with a new tensor of errors.

        Args:
            correct (```torch.Tensor``): A tensor that indicates the
                correct predictions. ``1`` indicates a correct
                prediction and ``0`` indicates a bad prediction.
        """
        self._meter.update(correct.detach())

    def value(self, prefix: str = "", suffix: str = "") -> dict[str, Union[int, float]]:
        meter = self._meter.all_reduce()
        if not meter.count:
            raise EmptyMetricError(f"{self.__class__.__qualname__} is empty")

        accuracy = meter.mean()
        num_correct_predictions = int(meter.sum())
        num_predictions = meter.count
        results = {
            f"{prefix}accuracy{suffix}": accuracy,
            f"{prefix}error{suffix}": 1.0 - accuracy,
            f"{prefix}num_correct_predictions{suffix}": num_correct_predictions,
            f"{prefix}num_incorrect_predictions{suffix}": num_predictions - num_correct_predictions,
            f"{prefix}num_predictions{suffix}": num_predictions,
        }
        for name, value in results.items():
            logger.info(f"{name}: {str_scalar(value)}")
        return results


def setup_state(state: Union[BaseState, dict]) -> BaseState:
    r"""Sets up the metric state.

    Args:
        state (``BaseState`` or dict): Specifies the metric state or
            its configuration.

    Returns:
        ``BaseState``: The instantiated metric state.
    """
    if isinstance(state, dict):
        logger.info(
            f"Initializing a metric state from its configuration... {str_target_object(state)}"
        )
        state = BaseState.factory(**state)
    return state



__all__ = [
    "SequentialIterDataPipeCreator",
    "SequentialCreatorIterDataPipeCreator",
    "create_sequential_iter_datapipe",
]

from collections.abc import Sequence
from typing import Optional, Union

from objectory import OBJECT_TARGET, factory
from torch.utils.data import IterDataPipe

from gravitorch.creators.datapipe.base import (
    BaseIterDataPipeCreator,
    setup_iter_datapipe_creator,
)
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_add_indent, to_torch_sequence_str


class SequentialIterDataPipeCreator(BaseIterDataPipeCreator):
    r"""Implements an ``IterDataPipe`` creator to create a sequence of
    ``IterDataPipe``s from their configuration.

    Args:
        config (dict or sequence of dict): Specifies the configuration
            of the ``IterDataPipe`` object to create. See description
            of the ``create_sequential_iter_datapipe`` function to
            learn more about the expected values.

    Raises:
        ValueError if the ``IterDataPipe`` configuration sequence is
            empty.
    """

    def __init__(self, config: Union[dict, Sequence[dict]]):
        if not config:
            raise ValueError("It is not possible to create a DataPipe because `config` is empty")
        self._config = config

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_add_indent(to_torch_sequence_str(self._config))},\n)"
        )

    def create(
        self, engine: Optional[BaseEngine] = None, source_inputs: Optional[Sequence] = None
    ) -> IterDataPipe:
        r"""Creates an ``IterDataPipe`` object.

        Args:
            engine (``BaseEngine`` or ``None``, optional): Specifies
                an engine. The engine is not used by this creator.
                Default: ``None``
            source_inputs (sequence or ``None``): Specifies the first
                positional arguments of the source ``IterDataPipe``.
                This argument can be used to create a new
                ``IterDataPipe`` object, that takes existing
                ``IterDataPipe`` objects as input. See examples below
                to see how to use it. If ``None``, ``source_inputs`` is
                set to an empty tuple. Default: ``None``

        Returns:
            ``IterDataPipe``: The created ``IterDataPipe`` object.

        Example usage:

        .. code-block:: python

            >>> from torch.utils.data import IterDataPipe
            >>> from objectory import OBJECT_TARGET
            >>> from gravitorch.creators.datapipe import SequentialIterDataPipeCreator
            # Create an IterDataPipe object using a single IterDataPipe object and no source input
            >>> creator = SequentialIterDataPipeCreator(
            ...     {
            ...         OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
            ...         "data": [1, 2, 3, 4],
            ...     }
            ... )
            >>> datapipe: IterDataPipe = creator.create()
            >>> tuple(datapipe)
            (1, 2, 3, 4)
            # Equivalent to
            >>> creator = SequentialIterDataPipeCreator(
            ...     [
            ...         {
            ...             OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
            ...             "data": [1, 2, 3, 4],
            ...         },
            ...     ]
            ... )
            >>> datapipe: IterDataPipe = creator.create()
            >>> tuple(datapipe)
            (1, 2, 3, 4)
            # It is possible to use the source_inputs to create the same IterDataPipe object.
            # The data is given by the source_inputs
            >>> creator = SequentialIterDataPipeCreator(
            ...     config={OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe"},
            ... )
            >>> datapipe: IterDataPipe = creator.create(source_inputs=([1, 2, 3, 4],))
            >>> tuple(datapipe)
            (1, 2, 3, 4)
            # Create an IterDataPipe object using two IterDataPipe objects and no source input
            >>> creator = SequentialIterDataPipeCreator(
            ...     [
            ...         {
            ...             OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
            ...             "data": [1, 2, 3, 4],
            ...         },
            ...         {
            ...             OBJECT_TARGET: "torch.utils.data.datapipes.iter.Batcher",
            ...             "batch_size": 2,
            ...         },
            ...     ]
            ... )
            >>> datapipe: IterDataPipe = creator.create()
            >>> tuple(datapipe)
            ([1, 2], [3, 4])
            # It is possible to use the source_inputs to create the same IterDataPipe object.
            # A source IterDataPipe object is specified by using source_inputs
            >>> creator = SequentialIterDataPipeCreator(
            ...     config=[
            ...         {
            ...             OBJECT_TARGET: "torch.utils.data.datapipes.iter.Batcher",
            ...             "batch_size": 2,
            ...         },
            ...     ],
            ... )
            >>> from gravitorch.data.datapipes.iter import SourceWrapperIterDataPipe
            >>> datapipe: IterDataPipe = creator.create(
            ...     source_inputs=[SourceWrapperIterDataPipe(data=[1, 2, 3, 4])]
            ... )
            >>> tuple(datapipe)
            ([1, 2], [3, 4])
            # It is possible to create a sequential ``IterDataPipe`` object that takes several
            # IterDataPipe objects as input.
            >>> creator = SequentialIterDataPipeCreator(
            ...     config=[
            ...         {OBJECT_TARGET: "torch.utils.data.datapipes.iter.Multiplexer"},
            ...         {
            ...             OBJECT_TARGET: "torch.utils.data.datapipes.iter.Batcher",
            ...             "batch_size": 2,
            ...         },
            ...     ],
            ... )
            >>> datapipe: IterDataPipe = creator.create(
            ...    source_inputs=[
            ...         SourceWrapperIterDataPipe(data=[1, 2, 3, 4]),
            ...         SourceWrapperIterDataPipe(data=[11, 12, 13, 14]),
            ...     ],
            ... )
            >>> tuple(datapipe)
            ([1, 11], [2, 12], [3, 13], [4, 14])
        """
        return create_sequential_iter_datapipe(config=self._config, source_inputs=source_inputs)


def create_sequential_iter_datapipe(
    config: Union[dict, Sequence[dict]],
    source_inputs: Optional[Sequence] = None,
) -> IterDataPipe:
    r"""Creates a sequential ``IterDataPipe`` object.

    A sequential ``IterDataPipe`` object has a single source (which
    can takes multiple ``IterDataPipe`` objects) and a single sink.
    The structure should look like:

        SourceDatapipe -> DataPipe1 -> DataPipe2 -> SinkDataPipe

    The structure of the ``config`` input depends on the sequential
    ``IterDataPipe`` object that is created:

        - If ``config`` is a ``dict`` object, it creates a sequential
            ``IterDataPipe`` object with a single ``IterDataPipe``
            object. The dictionary should contain the parameters used
            to initialize the ``IterDataPipe`` object. It should
            follow the ``object_factory`` syntax. Using a dict allows
            to initialize a single ``IterDataPipe`` object. If you
            want to create a ``IterDataPipe`` object recursively, you
            need to give a sequence of dict.
        - If ``config`` is a sequence of ``dict`` objects, this
            function creates an ``IterDataPipe`` object with a
            sequential structure. The sequence of configurations
            follows the order of the ``IterDataPipe``s. The first
            config is used to create the first ``IterDataPipe``
            (a.k.a. source), and the last config is used to create the
            last ``IterDataPipe`` (a.k.a. sink). This function assumes
            all the DataPipes have a single source DataPipe as their
            first argument, excepts for the source ``IterDataPipe``.

    Note: it is possible to create sequential ``IterDataPipe`` objects
    without using this function.

    Args:
        config (dict or sequence of dict): Specifies the configuration
            of the ``IterDataPipe`` object to create. See description
            above to know when to use a dict or a sequence of dicts.
        source_inputs (sequence or ``None``): Specifies the first
            positional arguments of the source ``IterDataPipe``. This
            argument can be used to create a new ``IterDataPipe``
            object, that takes existing ``IterDataPipe`` objects as
            input. See examples below to see how to use it.
            If ``None``, ``source_inputs`` is set to an empty tuple.
            Default: ``None``

    Returns:
        ``IterDataPipe``: The last (a.k.a. sink) ``IterDataPipe`` of
            the sequence.

    Raises:
        ValueError if the configuration is empty (empty dict or
            sequence).

    Example usage:

    .. code-block:: python

        >>> from torch.utils.data import IterDataPipe
        >>> from objectory import OBJECT_TARGET
        >>> from gravitorch.creators.datapipe.sequential import create_sequential_iter_datapipe
        # Create an IterDataPipe object using a single IterDataPipe object and no source input
        >>> datapipe: IterDataPipe = create_sequential_iter_datapipe(
        ...     {
        ...         OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
        ...         "data": [1, 2, 3, 4],
        ...     }
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        # Equivalent to
        >>> datapipe: IterDataPipe = create_sequential_iter_datapipe(
        ...     [
        ...         {
        ...             OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
        ...             "data": [1, 2, 3, 4],
        ...         },
        ...     ]
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        # It is possible to use the source_inputs to create the same IterDataPipe object.
        # The data is given by the source_inputs
        >>> datapipe: IterDataPipe = create_sequential_iter_datapipe(
        ...     config={OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe"},
        ...     source_inputs=([1, 2, 3, 4],),
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
        # Create an IterDataPipe object using two IterDataPipe objects and no source input
        >>> datapipe: IterDataPipe = create_sequential_iter_datapipe(
        ...     [
        ...         {
        ...             OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
        ...             "data": [1, 2, 3, 4],
        ...         },
        ...         {OBJECT_TARGET: "torch.utils.data.datapipes.iter.Batcher", "batch_size": 2},
        ...     ]
        ... )
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
        # It is possible to use the source_inputs to create the same IterDataPipe object.
        # A source IterDataPipe object is specified by using source_inputs
        >>> from gravitorch.data.datapipes.iter import SourceWrapperIterDataPipe
        >>> datapipe: IterDataPipe = create_sequential_iter_datapipe(
        ...     config=[
        ...         {OBJECT_TARGET: "torch.utils.data.datapipes.iter.Batcher", "batch_size": 2},
        ...     ],
        ...     source_inputs=[SourceWrapperIterDataPipe(data=[1, 2, 3, 4])],
        ... )
        >>> tuple(datapipe)
        ([1, 2], [3, 4])
        # It is possible to create a sequential ``IterDataPipe`` object that takes several
        # IterDataPipe objects as input.
        >>> datapipe: IterDataPipe = create_sequential_iter_datapipe(
        ...     config=[
        ...         {OBJECT_TARGET: "torch.utils.data.datapipes.iter.Multiplexer"},
        ...         {OBJECT_TARGET: "torch.utils.data.datapipes.iter.Batcher", "batch_size": 2},
        ...     ],
        ...     source_inputs=[
        ...         SourceWrapperIterDataPipe(data=[1, 2, 3, 4]),
        ...         SourceWrapperIterDataPipe(data=[11, 12, 13, 14]),
        ...     ],
        ... )
        >>> tuple(datapipe)
        ([1, 11], [2, 12], [3, 13], [4, 14])
    """
    if not config:
        raise ValueError("It is not possible to create a DataPipe because `config` is empty")
    source_inputs = source_inputs or tuple()
    if isinstance(config, dict):
        config = config.copy()  # Make a copy because the dict is modified below.
        target = config.pop(OBJECT_TARGET)
        return factory(target, *source_inputs, **config)
    datapipe = create_sequential_iter_datapipe(config[0], source_inputs)
    for cfg in config[1:]:
        datapipe = create_sequential_iter_datapipe(cfg, source_inputs=(datapipe,))
    return datapipe


class SequentialCreatorIterDataPipeCreator(BaseIterDataPipeCreator):
    r"""Implements an ``IterDataPipe`` creator to create an ``IterDataPipe``
    object by using a sequence ``IterDataPipe`` creators.

    Args:
        creators: Specifies the sequence of ``IterDataPipe`` creators
            or their configurations. The sequence of creators follows
            the order of the ``IterDataPipe``s. The first creator is
            used to create the first ``IterDataPipe`` (a.k.a. source),
            and the last creator is used to create the last
            ``IterDataPipe`` (a.k.a. sink). This creator assumes all
            the DataPipes have a single source DataPipe as their first
            argument, excepts for the source ``IterDataPipe``.
    """

    def __init__(self, creators: Sequence[Union[BaseIterDataPipeCreator, dict]]):
        if not creators:
            raise ValueError("It is not possible to create a DataPipe because `creators` is empty")
        self._creators = [setup_iter_datapipe_creator(creator) for creator in creators]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  {str_add_indent(to_torch_sequence_str(self._creators))},\n)"
        )

    def create(
        self, engine: Optional[BaseEngine] = None, source_inputs: Optional[Sequence] = None
    ) -> IterDataPipe:
        r"""Creates an ``IterDataPipe`` object.

        Args:
            engine (``BaseEngine`` or ``None``, optional): Specifies
                an engine.
                The engine is not used by this creator.
                Default: ``None``
            source_inputs (sequence or ``None``): Specifies the first
                positional arguments of the source ``IterDataPipe``.
                This argument can be used to create a new
                ``IterDataPipe`` object, that takes existing
                ``IterDataPipe`` objects as input.  See examples below
                to see how to use it. If ``None``, ``source_inputs``
                is setto an empty tuple. Default: ``None``

        Returns:
            ``IterDataPipe``: The created ``IterDataPipe`` object.

        Example usage:

        .. code-block:: python

            >>> from torch.utils.data import IterDataPipe
            >>> from objectory import OBJECT_TARGET
            >>> from gravitorch.creators.datapipe import (
            ...     SequentialCreatorIterDataPipeCreator,
            ...     SequentialIterDataPipeCreator,
            ... )
            # Create an IterDataPipe object using a single IterDataPipe creator and no source input
            >>> creator = SequentialCreatorIterDataPipeCreator(
            ...     [
            ...         SequentialIterDataPipeCreator(
            ...             {
            ...                 OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe",
            ...                 "data": [1, 2, 3, 4],
            ...             },
            ...         ),
            ...     ]
            ... )
            >>> datapipe: IterDataPipe = creator.create()
            >>> tuple(datapipe)
            (1, 2, 3, 4)
            # It is possible to use the source_inputs to create the same IterDataPipe object.
            # The data is given by the source_inputs
            >>> creator = SequentialCreatorIterDataPipeCreator(
            ...     [
            ...         SequentialIterDataPipeCreator(
            ...             {OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe"},
            ...         ),
            ...     ]
            ... )
            >>> datapipe: IterDataPipe = creator.create(source_inputs=([1, 2, 3, 4],))
            >>> tuple(datapipe)
            (1, 2, 3, 4)
            # Create an IterDataPipe object using two IterDataPipe creators and no source input
            >>> creator = SequentialCreatorIterDataPipeCreator(
            ...     [
            ...         SequentialIterDataPipeCreator(
            ...             {OBJECT_TARGET: "gravitorch.data.datapipes.iter.SourceWrapperIterDataPipe"},
            ...         ),
            ...         SequentialIterDataPipeCreator(
            ...             {
            ...                 OBJECT_TARGET: "torch.utils.data.datapipes.iter.Batcher",
            ...                 "batch_size": 2,
            ...             },
            ...         ),
            ...     ]
            ... )
            >>> datapipe: IterDataPipe = creator.create()
            >>> tuple(datapipe)
            ([1, 2], [3, 4])
            # It is possible to use the source_inputs to create the same IterDataPipe object.
            # A source IterDataPipe object is specified by using source_inputs
            >>> from gravitorch.data.datapipes.iter import SourceWrapperIterDataPipe
            >>> creator = SequentialCreatorIterDataPipeCreator(
            ...     creators=[
            ...         SequentialIterDataPipeCreator(
            ...             {
            ...                 OBJECT_TARGET: "torch.utils.data.datapipes.iter.Batcher",
            ...                 "batch_size": 2,
            ...             },
            ...         ),
            ...     ]
            ... )
            >>> datapipe: IterDataPipe = creator.create(
            ...     source_inputs=[SourceWrapperIterDataPipe(data=[1, 2, 3, 4])]
            ... )
            >>> tuple(datapipe)
            ([1, 2], [3, 4])
            # It is possible to create a sequential ``IterDataPipe`` object that takes several
            # IterDataPipe objects as input.
            >>> creator = SequentialCreatorIterDataPipeCreator(
            ...     [
            ...         SequentialIterDataPipeCreator(
            ...             {OBJECT_TARGET: "torch.utils.data.datapipes.iter.Multiplexer"},
            ...         ),
            ...         SequentialIterDataPipeCreator(
            ...             {
            ...                 OBJECT_TARGET: "torch.utils.data.datapipes.iter.Batcher",
            ...                 "batch_size": 2,
            ...             },
            ...         ),
            ...     ]
            ... )
            >>> datapipe: IterDataPipe = creator.create(
            ...    source_inputs=[
            ...         SourceWrapperIterDataPipe(data=[1, 2, 3, 4]),
            ...         SourceWrapperIterDataPipe(data=[11, 12, 13, 14]),
            ...     ],
            ... )
            >>> tuple(datapipe)
            ([1, 11], [2, 12], [3, 13], [4, 14])
        """
        datapipe = self._creators[0].create(engine=engine, source_inputs=source_inputs)
        for creator in self._creators[1:]:
            datapipe = creator.create(engine=engine, source_inputs=(datapipe,))
        return datapipe



r"""This module implements a training loop using the Accelerate library
(https://huggingface.co/docs/accelerate)."""

__all__ = ["AccelerateTrainingLoop"]

import logging
import sys
from collections.abc import Callable, Iterable
from typing import Any, Optional, Union

import torch
from torch.nn import Module
from torch.optim import Optimizer
from tqdm import tqdm

from gravitorch import constants as ct
from gravitorch.distributed import comm as dist
from gravitorch.engines.base import BaseEngine
from gravitorch.engines.events import EngineEvents
from gravitorch.utils.integrations import check_accelerate, is_accelerate_available
from gravitorch.utils.loop_observers import BaseLoopObserver
from gravitorch.utils.profilers import BaseProfiler
from gravitorch.utils.training_loops.base import BaseBasicTrainingLoop

if is_accelerate_available():
    from accelerate import Accelerator
else:
    Accelerator = None  # pragma: no cover

logger = logging.getLogger(__name__)


class AccelerateTrainingLoop(BaseBasicTrainingLoop):
    r"""Implements a training loop that uses ``accelerate.Accelerator`` to train
    a model.

    Args:
        accelerator (``accelerate.Accelerator`` or dict or None,
            optional): Specifies the ``accelerate.Accelerator`` object
            or the parameters to instantiate it. Please read the
            ``accelerate.Accelerator`` documentation to know the
            parameters
            https://huggingface.co/docs/accelerate/accelerator.html.
            If ``None``, it will use the default parameters.
            Default: ``None``
        set_grad_to_none (bool, optional): If ``True``, set the
            gradients to ``None``, otherwise set the gradients to
            zero. Setting the gradients to ``None`` will in general
            have lower memory footprint, and can modestly improve
            performance. Default: ``False``
        tag (str, optional): Specifies the tag which is used to log
            metrics. Default: ``"train"``
        clip_grad (dict or None, optional): Specifies the
            configuration to clip the gradient. If ``None``, no
            gradient clipping is used during the training.
            Default: ``None``
        observer (``BaseLoopObserver`` or dict or None, optional):
            Specifies the loop observer or its configuration.
            If ``None``, the ``NoOpLoopObserver`` is instantiated.
            Default: ``None``
        profiler (``BaseProfiler`` or dict or None, optional):
            Specifies the profiler or its configuration. If ``None``,
            the ``NoOpProfiler`` is instantiated. Default: ``None``
    """

    def __init__(
        self,
        accelerator: Union[Accelerator, dict, None] = None,
        set_grad_to_none: bool = False,
        tag: str = "train",
        clip_grad: Optional[dict] = None,
        observer: Union[BaseLoopObserver, dict, None] = None,
        profiler: Union[BaseProfiler, dict, None] = None,
    ):
        check_accelerate()
        self._accelerator = self._setup_accelerator(accelerator or {})
        logger.info(f"accelerator state:\n{self._accelerator.state}")
        super().__init__(tag=tag, clip_grad=clip_grad, observer=observer, profiler=profiler)
        self._set_grad_to_none = bool(set_grad_to_none)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  accelerator={self._accelerator},\n"
            f"  set_grad_to_none={self._set_grad_to_none},\n"
            f"  tag={self._tag},\n"
            f"  clip_grad_fn={self._clip_grad_fn},\n"
            f"  clip_grad_args={self._clip_grad_args},\n"
            f"  observer={self._observer},\n"
            f"  profiler={self._profiler},\n"
            ")"
        )

    def _prepare_model_optimizer_data_loader(
        self, engine: BaseEngine
    ) -> tuple[Module, Optimizer, Iterable]:
        r"""Prepares the model, optimizer and data loader.

        Args:
            engine (``BaseEngine``): Specifies the engine.

        Returns:
            ``torch.nn.Module``, ``torch.optim.Optimizer``,
                ``Iterable``: A tuple with the model, the optimizer
                and the data loader.
        """
        logger.info("Preparing the model, optimizer, and data loader...")
        model, optimizer, data_loader = self._accelerator.prepare(
            [
                engine.model,
                engine.optimizer,
                engine.data_source.get_data_loader(loader_id=self._tag, engine=engine),
            ],
        )
        prefix = f"({dist.get_rank()}/{dist.get_world_size()}) " if dist.is_distributed() else ""
        data_loader = tqdm(
            data_loader,
            desc=f"{prefix}Training [{engine.epoch}/{engine.max_epochs}]",
            position=dist.get_rank(),
            file=sys.stdout,
        )
        logger.info("Training data loader has been created")
        return model, optimizer, data_loader

    def _train_one_batch(
        self, engine: BaseEngine, model: Module, optimizer: Optimizer, batch: Any
    ) -> dict:
        engine.fire_event(EngineEvents.TRAIN_ITERATION_STARTED)
        optimizer.zero_grad(self._set_grad_to_none)
        output = model(batch)
        engine.fire_event(EngineEvents.TRAIN_FORWARD_COMPLETED)

        if not torch.isnan(output[ct.LOSS]):
            self._accelerator.backward(output[ct.LOSS])
        else:
            logger.warning(
                "NaN detected. The gradient is not computed for this batch "
                f"(iteration: {engine.iteration})"
            )

        if self._clip_grad_fn:
            self._clip_grad_fn(model.parameters(), *self._clip_grad_args)
        engine.fire_event(EngineEvents.TRAIN_BACKWARD_COMPLETED)

        optimizer.step()
        engine.fire_event(EngineEvents.TRAIN_ITERATION_COMPLETED)

        return output

    def _setup_accelerator(self, accelerator: Union[Accelerator, dict]) -> Accelerator:
        r"""Sets up the accelerator.

        Args:
            accelerator (``accelerate.Accelerator`` or dict, optional):
                Specifies the ``accelerate.Accelerator`` object or the
                parameters to instantiate it. Please read the
                ``accelerate.Accelerator`` documentation to know the
                parameters https://huggingface.co/docs/accelerate/accelerator.html.

        Returns:
            ``accelerate.Accelerator``: The accelerator object.

        Raises:
            RuntimeError: if the accelerate package is not installed.
        """
        if isinstance(accelerator, Accelerator):
            return accelerator
        logger.info(f"accelerator options: {accelerator}")
        return Accelerator(**accelerator)

    def _setup_clip_grad(self, clip_grad: dict) -> tuple[Optional[Callable], tuple]:
        if not clip_grad:
            return None, ()

        name = clip_grad["name"]
        if name == "clip_grad_value":
            clip_grad_fn = self._accelerator.clip_grad_value_
            clip_grad_args = (clip_grad.get("clip_value", 0.25),)
            logger.info(f"clip gradient by value {clip_grad_args[0]}")
            return clip_grad_fn, clip_grad_args
        if name == "clip_grad_norm":
            clip_grad_fn = self._accelerator.clip_grad_norm_
            clip_grad_args = clip_grad.get("max_norm", 1), clip_grad.get("norm_type", 2)
            logger.info(
                f"clip gradient by maximum norm {clip_grad_args[0]} (norm type: "
                f"{clip_grad_args[1]})"
            )
            return clip_grad_fn, clip_grad_args
        raise ValueError(
            f"Incorrect clip grad name ({name}). The valid values are ``clip_grad_value`` "
            "and ``clip_grad_norm``"
        )
